# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Classes for the PHPP 'Heating' (Annual Heating Energy Demand) Worksheet."""

from __future__ import annotations

from typing import Any, Dict

from ph_units.unit_type import Unit

from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_app


class HeatingDemand:
    """IO Controller for the PHPP 'Heating' (Annual Heating Energy Demand) Worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.HeatingDemand) -> None:
        self.xl = _xl
        self.shape = _shape

    def _get_annual_demand(self, _col: str) -> Dict[str, Unit]:
        shp = self.shape
        qT = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_total_losses_transmission}")
        qV = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_total_losses_ventilation}")
        qL = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_total_losses}")
        qS = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_total_gains_solar}")
        qI = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_total_gains_internal}")
        util = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_utilization_factor}")
        qG = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_useful_gains}")
        qH = self.xl.get_single_data_item(shp.name, f"{_col}{shp.row_annual_demand}")

        return {
            "losses_transmission": Unit(float(qT or 0.0), str(shp.unit)),
            "losses_ventilation": Unit(float(qV or 0.0), str(shp.unit)),
            "gains_solar": Unit(float(qS or 0.0), str(shp.unit)),
            "gains_internal": Unit(float(qI or 0.0), str(shp.unit)),
            "utilization_factor": Unit(float(util or 0.0), "%"),
            "heating_demand": Unit(float(qH or 0.0), str(shp.unit)),
        }

    def get_annual_demand(self) -> Dict[str, Unit]:
        """Return a Dict of all the Heating Energy Demand data (kWh-yr)."""
        return self._get_annual_demand(self.shape.col_kWh_year)

    def get_specific_annual_demand(self) -> Dict[str, Unit]:
        """Return a Dict of all the Specific Heating Energy Demand data (kWh/m2-yr)."""
        return self._get_annual_demand(self.shape.col_kWh_m2_year)



"""PHX Mechanical Equipment (water, heating, heat-pumps (cooling), ventilation)"""

from PHX.model.enums.hvac import *
from PHX.model.hvac._base import *
from PHX.model.hvac.collection import *
from PHX.model.hvac.cooling_params import *
from PHX.model.hvac.ducting import *
from PHX.model.hvac.heat_pumps import *
from PHX.model.hvac.heating import *
from PHX.model.hvac.piping import *
from PHX.model.hvac.renewable_devices import *
from PHX.model.hvac.supportive_devices import *
from PHX.model.hvac.ventilation import *
from PHX.model.hvac.water import *



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Mechanical Heat-Pump (Heating + Cooling) Devices."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, Union

from PHX.model.enums.hvac import DeviceType, HeatPumpType, SystemType
from PHX.model.hvac import _base
from PHX.model.hvac.cooling_params import PhxCoolingParams

# -----------------------------------------------------------------------------
# Base


@dataclass
class PhxHeatPumpDevice(_base.PhxMechanicalDevice):
    params_cooling: PhxCoolingParams = field(default_factory=PhxCoolingParams)

    def __post_init__(self) -> None:
        super().__post_init__()


# -----------------------------------------------------------------------------
# Params


@dataclass
class PhxHeatPumpAnnualParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.ANNUAL)
    annual_COP: Optional[float] = None
    total_system_perf_ratio: Optional[float] = None


@dataclass
class PhxHeatPumpMonthlyParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.RATED_MONTHLY)
    _COP_1: Optional[float] = None
    _COP_2: Optional[float] = None
    _ambient_temp_1: Optional[float] = None
    _ambient_temp_2: Optional[float] = None

    @property
    def monthly_COPS(self):
        return None

    @monthly_COPS.setter
    def monthly_COPS(self, _in):
        if not _in:
            return

        self.COP_1 = _in[0]
        try:
            self.COP_2 = _in[1]
        except IndexError:
            self.COP_2 = _in[0]

    @property
    def monthly_temps(self):
        return None

    @monthly_temps.setter
    def monthly_temps(self, _in):
        if not _in:
            return

        self.ambient_temp_1 = _in[0]
        try:
            self.ambient_temp_2 = _in[1]
        except IndexError:
            self.ambient_temp_2 = _in[0]

    @property
    def COP_1(self) -> Optional[float]:
        return self._COP_1

    @COP_1.setter
    def COP_1(self, value: Optional[float]) -> None:
        if value is not None:
            self._COP_1 = value

    @property
    def COP_2(self) -> Optional[float]:
        return self._COP_2

    @COP_2.setter
    def COP_2(self, value: Optional[float]) -> None:
        if value is not None:
            self._COP_2 = value

    @property
    def ambient_temp_1(self) -> Optional[float]:
        return self._ambient_temp_1

    @ambient_temp_1.setter
    def ambient_temp_1(self, value: Optional[float]) -> None:
        if value is not None:
            self._ambient_temp_1 = value

    @property
    def ambient_temp_2(self) -> Optional[float]:
        return self._ambient_temp_2

    @ambient_temp_2.setter
    def ambient_temp_2(self, value: Optional[float]) -> None:
        if value is not None:
            self._ambient_temp_2 = value


@dataclass
class PhxHeatPumpHotWaterParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.HOT_WATER)
    annual_COP: Optional[float] = None
    total_system_perf_ratio: Optional[float] = None
    annual_energy_factor: Optional[float] = None


@dataclass
class PhxHeatPumpCombinedParams(_base.PhxMechanicalDeviceParams):
    hp_type: HeatPumpType = field(init=False, default=HeatPumpType.COMBINED)


AnyHeatPumpParams = Union[
    PhxHeatPumpAnnualParams,
    PhxHeatPumpMonthlyParams,
    PhxHeatPumpHotWaterParams,
    PhxHeatPumpCombinedParams,
]


# -----------------------------------------------------------------------------
# Equipment


@dataclass
class PhxHeatPumpAnnual(PhxHeatPumpDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeatPumpAnnualParams = field(default_factory=PhxHeatPumpAnnualParams)
    params_cooling: PhxCoolingParams = field(default_factory=PhxCoolingParams)


@dataclass
class PhxHeatPumpMonthly(PhxHeatPumpDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeatPumpMonthlyParams = field(default_factory=PhxHeatPumpMonthlyParams)
    params_cooling: PhxCoolingParams = field(default_factory=PhxCoolingParams)


@dataclass
class PhxHeatPumpCombined(PhxHeatPumpDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeatPumpCombinedParams = field(default_factory=PhxHeatPumpCombinedParams)
    params_cooling: PhxCoolingParams = field(default_factory=PhxCoolingParams)


@dataclass
class PhxHeatPumpHotWater(PhxHeatPumpDevice):
    system_type: SystemType = field(init=False, default=SystemType.HEAT_PUMP)
    device_type: DeviceType = field(init=False, default=DeviceType.HEAT_PUMP)
    params: PhxHeatPumpHotWaterParams = field(default_factory=PhxHeatPumpHotWaterParams)
    params_cooling: PhxCoolingParams = field(default_factory=PhxCoolingParams)


AnyPhxHeatPump = Union[
    PhxHeatPumpAnnual,
    PhxHeatPumpMonthly,
    PhxHeatPumpHotWater,
    PhxHeatPumpCombined,
]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Program Class for organizing Occupancy Data (People)"""

from __future__ import annotations

from dataclasses import dataclass, field

from PHX.model.loads.occupancy import PhxLoadOccupancy
from PHX.model.schedules.occupancy import PhxScheduleOccupancy


@dataclass
class PhxProgramOccupancy:
    """A PHX Program for the Occupancy (People) with a load and schedule."""

    display_name: str = "Unnamed_Occupancy_Program"
    load: PhxLoadOccupancy = field(default_factory=PhxLoadOccupancy)
    schedule: PhxScheduleOccupancy = field(default_factory=PhxScheduleOccupancy)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Passive House Renewable Energy (PV) Device classes"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from PHX.model.enums.hvac import DeviceType, SystemType
from PHX.model.hvac import _base


@dataclass
class PhxDevicePhotovoltaicParams(_base.PhxMechanicalDeviceParams):
    """PV System Parameters"""

    _location_type: int = 1
    _onsite_utilization_type: int = 2
    _utilization_type: int = 1

    _array_size: float = 0.0
    _photovoltaic_renewable_energy: float = 0.0
    _onsite_utilization_factor: float = 1.0
    _auxiliary_energy: float = 0.0
    _auxiliary_energy_DHW: float = 0.0
    _in_conditioned_space: bool = False

    @property
    def location_type(self) -> int:
        return self._location_type

    @location_type.setter
    def location_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._location_type = value

    @property
    def onsite_utilization_type(self) -> int:
        return self._onsite_utilization_type

    @onsite_utilization_type.setter
    def onsite_utilization_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._onsite_utilization_type = value

    @property
    def utilization_type(self) -> int:
        return self._utilization_type

    @utilization_type.setter
    def utilization_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._utilization_type = value

    @property
    def array_size(self) -> float:
        return self._array_size

    @array_size.setter
    def array_size(self, value: Optional[float]) -> None:
        if value is not None:
            self._array_size = value

    @property
    def photovoltaic_renewable_energy(self) -> float:
        return self._photovoltaic_renewable_energy

    @photovoltaic_renewable_energy.setter
    def photovoltaic_renewable_energy(self, value: Optional[float]) -> None:
        if value is not None:
            self._photovoltaic_renewable_energy = value

    @property
    def onsite_utilization_factor(self) -> float:
        return self._onsite_utilization_factor

    @onsite_utilization_factor.setter
    def onsite_utilization_factor(self, value: Optional[float]) -> None:
        if value is not None:
            self._onsite_utilization_factor = value

    @property
    def auxiliary_energy(self) -> float:
        return self._auxiliary_energy

    @auxiliary_energy.setter
    def auxiliary_energy(self, value: Optional[float]) -> None:
        if value is not None:
            self._auxiliary_energy = value

    @property
    def auxiliary_energy_DHW(self) -> float:
        return self._auxiliary_energy_DHW

    @auxiliary_energy_DHW.setter
    def auxiliary_energy_DHW(self, value: Optional[float]) -> None:
        if value is not None:
            self._auxiliary_energy_DHW = value

    @property
    def in_conditioned_space(self) -> bool:
        return self._in_conditioned_space

    @in_conditioned_space.setter
    def in_conditioned_space(self, value: Optional[bool]) -> None:
        if value is not None:
            self._in_conditioned_space = value

    def __add__(self, other: PhxDevicePhotovoltaicParams) -> PhxDevicePhotovoltaicParams:
        return PhxDevicePhotovoltaicParams(
            _location_type=self.location_type,
            _onsite_utilization_type=self.onsite_utilization_type,
            _utilization_type=self.utilization_type,
            _array_size=self.array_size + other.array_size,
            _photovoltaic_renewable_energy=self.photovoltaic_renewable_energy + other.photovoltaic_renewable_energy,
            _onsite_utilization_factor=(self.onsite_utilization_factor + other.onsite_utilization_factor) / 2,
            _auxiliary_energy=self.auxiliary_energy + other.auxiliary_energy,
            _auxiliary_energy_DHW=self.auxiliary_energy_DHW + other.auxiliary_energy_DHW,
            in_conditioned_space=self.in_conditioned_space,
        )


@dataclass
class PhxDevicePhotovoltaic(_base.PhxMechanicalDevice):
    """PV Panel Systems."""

    system_type: SystemType = field(init=False, default=SystemType.PHOTOVOLTAIC)
    device_type: DeviceType = field(init=False, default=DeviceType.PHOTOVOLTAIC)
    params: PhxDevicePhotovoltaicParams = field(default_factory=PhxDevicePhotovoltaicParams)

    def __post_init__(self) -> None:
        super().__post_init__()

    def __add__(self, other: PhxDevicePhotovoltaic) -> PhxDevicePhotovoltaic:
        if self.system_type.value != other.system_type.value:
            raise ValueError(f"Error: Cannot add {self.system_type.value} to {other.system_type.value}")
        if self.device_type.value != other.device_type.value:
            raise ValueError(f"Error: Cannot add {self.device_type.value} to {other.device_type.value}")
        return self + other


AnyRenewableDevice = PhxDevicePhotovoltaic



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP Windows worksheet."""

from __future__ import annotations

from typing import Generator, List, Optional

from ph_units.unit_type import Unit

from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model.windows_rows import WindowRow, get_name_from_glazing_id
from PHX.xl import xl_app
from PHX.xl.xl_data import XlItem, col_offset


class Windows:
    """IO Controller Class for PHPP "Windows" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Windows):
        self.xl = _xl
        self.shape = shape
        self._header_row: Optional[int] = None
        self._first_entry_row: Optional[int] = None
        self._last_entry_row: Optional[int] = None

    @property
    def header_row(self) -> int:
        """The row number for the Window entry 'Header'."""
        if not self._header_row:
            self._header_row = self.find_header_row()
        return self._header_row

    @property
    def first_entry_row(self) -> int:
        """Return the starting row for the window data entry block."""
        if not self._first_entry_row:
            self._first_entry_row = self.find_first_entry_row()
        return self._first_entry_row

    @property
    def last_entry_row(self) -> int:
        """Return the ending row for the window data entry block."""
        if not self._last_entry_row:
            self._last_entry_row = self.find_last_entry_row()
        return self._last_entry_row

    @property
    def entry_range_start(self) -> str:
        return f"{self.shape.window_rows.locator_col_entry}{self.first_entry_row}"

    @property
    def entry_range_end(self) -> str:
        return f"{self.shape.window_rows.locator_col_entry}{self.last_entry_row}"

    @property
    def entry_range(self) -> str:
        """Return the range for the window data entry block."""
        return f"{self.entry_range_start}:{self.entry_range_end}"

    @property
    def windows_row_numbers(self) -> List[int]:
        """Return a list of all the window row numbers."""
        return list(range(self.first_entry_row, self.last_entry_row + 1))

    @property
    def used_window_row_numbers(self) -> Generator[int, None, None]:
        def has_window_data(_item) -> bool:
            if _item == "-" or _item == "<End of designPH import!>" or _item is None:
                return False
            else:
                return True

        data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.window_rows.inputs.description.column),
            self.first_entry_row,
            self.last_entry_row,
        )
        return (i for i, val in enumerate(data, start=self.first_entry_row) if has_window_data(val))

    def find_header_row(self, _row_start: int = 1, _read_length: int = 100) -> int:
        """Return the row number for the Window entry section 'Header'"""

        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_start + _read_length,
        )

        for i, val in enumerate(xl_data):
            if self.shape.window_rows.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.window_rows.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.window_rows.locator_string_header}?"
        )

    def find_first_entry_row(self, _start_row: int = 1, _read_length: int = 100) -> int:
        """Return the starting row for the window data entry block."""

        end_row = _start_row + _read_length
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows.locator_col_entry,
            _row_start=_start_row,
            _row_end=end_row,
        )

        for i, val in enumerate(xl_data, start=_start_row):
            if self.shape.window_rows.locator_string_entry in str(val):
                return i + 2

        if end_row < 10_000:
            return self.find_first_entry_row(_start_row=end_row, _read_length=1_000)

        raise Exception(
            f"Error: Cannot find the '{self.shape.window_rows.locator_string_entry}' "
            f"marker on the '{self.shape.name}' sheet, column {self.shape.window_rows.locator_col_entry}?"
        )

    def find_last_entry_row(self, _start_row: Optional[int] = None) -> int:
        """Return the last row of the Window input section."""
        if not _start_row:
            _start_row = self.first_entry_row
        elif _start_row > 10_000:
            raise Exception(
                f"Error: Cannot find the last row in the '{self.shape.name}' sheet, column {self.shape.window_rows_end.locator_col_entry}?"
            )

        _row_end = _start_row + 500
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows_end.locator_col_entry,
            _row_start=_start_row,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_start_row):
            if self.shape.window_rows_end.locator_string_entry in str(val):
                return i + 2
        else:
            return self.find_last_entry_row(_row_end)

    def set_single_window_construction_ids(
        self,
        _row_num: int,
        _glazing_construction_id: str,
        _frame_construction_id: str,
    ) -> None:
        """Set the glazing and frame construction IDs for a single window.

        Arguments:
        ---------
            * _row_num: int
                The row number for the window to set the construction IDs for.
            * _glazing_construction_id: str
                The glazing construction ID to set.
            * _frame_construction_id: str
                The frame construction ID to set.

        Returns:
        --------
            * None
        """
        glazing_col = str(self.shape.window_rows.inputs.glazing_id.column)
        glazing_range = f"{glazing_col}{_row_num}"
        self.xl.write_xl_item(XlItem(self.shape.name, glazing_range, _glazing_construction_id))

        frame_col = str(self.shape.window_rows.inputs.frame_id.column)
        frame_range = f"{frame_col}{_row_num}"
        self.xl.write_xl_item(XlItem(self.shape.name, frame_range, _frame_construction_id))

    def write_single_window(self, _row_num: int, _window_row: WindowRow) -> None:
        """Write a single WindowRow object to the Windows worksheet."""
        for item in _window_row.create_xl_items(self.shape.name, _row_num=_row_num):
            self.xl.write_xl_item(item)

    def write_windows(self, _window_rows: List[WindowRow]) -> None:
        """Write a list of WindowRow objects to the Windows worksheet."""
        for i, window_row in enumerate(_window_rows, start=self.first_entry_row):
            self.write_single_window(i, window_row)

    def get_all_window_names(self) -> List[str]:
        """Return a list of all the window names found in the worksheet."""

        # -- Get all the window names from the description row
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.window_rows.inputs.description.column,  # type: ignore
            _row_start=self.first_entry_row,
            _row_end=self.last_entry_row,
        )

        return [str(_) for _ in xl_data if _]

    def get_total_window_area(self, _tolerance: float = 1.0) -> Unit:
        """Return the total window area from the PHPP Windows worksheet."""
        angle_data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.window_rows.inputs.vertical_angle.column),
            self.first_entry_row,
            self.last_entry_row,
        )
        area_data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.window_rows.inputs.window_area.column),
            self.first_entry_row,
            self.last_entry_row,
        )

        areas = []
        for angle, area in zip(angle_data, area_data):
            if angle is None or area is None:
                continue
            if abs(90 - float(angle)) < _tolerance:
                areas.append(float(area))

        return Unit(sum(areas), str(self.shape.window_rows.inputs.window_area.unit))

    def get_total_skylight_area(self, _tolerance: float = 1.0) -> Unit:
        """Return the total skylight area from the PHPP Windows worksheet."""
        angle_data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.window_rows.inputs.vertical_angle.column),
            self.first_entry_row,
            self.last_entry_row,
        )
        area_data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.window_rows.inputs.window_area.column),
            self.first_entry_row,
            self.last_entry_row,
        )

        areas = []
        for angle, area in zip(angle_data, area_data):
            if angle is None or area is None:
                continue
            if abs(90 - float(angle)) > _tolerance:
                areas.append(float(area))

        return Unit(sum(areas), str(self.shape.window_rows.inputs.window_area.unit))

    def get_all_glazing_names(self) -> set[str]:
        """Return a set of all the construction names used in the Areas worksheet."""
        glazing_data = self.xl.get_single_column_data(
            self.shape.name,
            str(self.shape.window_rows.inputs.glazing_id.column),
            self.first_entry_row,
            self.last_entry_row,
        )
        return {get_name_from_glazing_id(_d) for _d in glazing_data}

    def activate_variants(self) -> None:
        """Set the frame and glass values to link to the Variants worksheet."""

        for row_num in range(self.first_entry_row, self.last_entry_row):
            # -- Link Glazing to the variants type
            self.xl.write_xl_item(
                XlItem(
                    self.shape.name,
                    f"{self.shape.window_rows.inputs.glazing_id.column}{row_num}",
                    f"={col_offset(str(self.shape.window_rows.inputs.variant_input.column), 1)}{row_num}",
                )
            )

            # -- Link Frame to the variants type
            self.xl.write_xl_item(
                XlItem(
                    self.shape.name,
                    f"{self.shape.window_rows.inputs.frame_id.column}{row_num}",
                    f"={col_offset(str(self.shape.window_rows.inputs.variant_input.column), 2)}{row_num}",
                )
            )

    def scale_window_size(self, _row_num: int, _scale_factor: float) -> None:
        """Scale the size of a single window based on an overall scale-factor.

        This is used during baseline model generation to scale the size of the windows
        in order to set the total WWR to the specified code-required value (40%).

        Arguments:
        ----------
            * _row_num: int
                The row number of the window to scale.
            * _scale_factor: float
                The factor to scale the window by.
        """

        # -- Current dimensions of the rectangle
        # -- Get the length and width from the PHPP Windows worksheet
        width = self.xl.get_data(
            self.shape.name,
            f"{self.shape.window_rows.inputs.width.column}{_row_num}",
        )
        height = self.xl.get_data(
            self.shape.name,
            f"{self.shape.window_rows.inputs.height.column}{_row_num}",
        )

        # -- Calculate current area
        current_area = float(height) * float(width)  # type: ignore
        desired_area = current_area * _scale_factor

        # -- Calculate scaling factor
        # ** 0.5 is equivalent to taking the square root of the expression on the
        # left-hand side. This is because raising a number to the power of 1/2 is
        # the same as taking the square root of that number.
        edge_scaling_factor = (desired_area / current_area) ** 0.5

        # -- Scale the dimensions
        new_width = width * edge_scaling_factor
        new_height = height * edge_scaling_factor

        # -- Set the new dimensions
        self.xl.write_xl_item(
            XlItem(
                self.shape.name,
                f"{self.shape.window_rows.inputs.width.column}{_row_num}",
                new_width,
            )
        )
        self.xl.write_xl_item(
            XlItem(
                self.shape.name,
                f"{self.shape.window_rows.inputs.height.column}{_row_num}",
                new_height,
            )
        )

    def row_is_window(self, _row_num: int, _tolerance: float = 5.0) -> bool:
        """Return True if the row is a window, False otherwise."""
        col = str(self.shape.window_rows.inputs.vertical_angle.column)
        _range = f"{col}{_row_num}"
        v = self.xl.get_data(self.shape.name, _range)

        if abs(90.0 - float(v)) < _tolerance:
            return True
        else:
            return False

    def row_is_skylight(self, _row_num: int, _tolerance: float = 5.0) -> bool:
        """Return True if the row is a skylight, False otherwise."""
        col = str(self.shape.window_rows.inputs.vertical_angle.column)
        _range = f"{col}{_row_num}"
        v = self.xl.get_data(self.shape.name, _range)

        if abs(90.0 - float(v)) < _tolerance:
            return False
        else:
            return True



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions to create new Shade PhxComponents from HB-Model Orphaned-Shade Objects."""

import operator
from collections import defaultdict
from functools import reduce
from typing import List

try:
    from honeybee import model
    from honeybee.shade import Shade
except ImportError as e:
    raise ImportError("\nFailed to import Honeybee:\n\t{}".format(e))

try:
    from PHX.from_HBJSON import create_geometry
    from PHX.from_HBJSON.cleanup_merge_faces import merge_hb_shades
    from PHX.model import project
    from PHX.model.components import PhxComponentOpaque
    from PHX.model.enums.building import ComponentColor, ComponentExposureExterior, ComponentFaceOpacity
except ImportError as e:
    raise ImportError("\nFailed to import PHX:\n\t{}".format(e))

try:
    from honeybee_ph_utils import face_tools
except ImportError as e:
    raise ImportError("\nFailed to import honeybee_ph_utils:\n\t{}".format(e))


def create_new_component_from_orphaned_shade(
    _shade: Shade,
) -> PhxComponentOpaque:
    """Returns a new PHX-Component for based on the Honeybee-Shade.

    Arguments:
    ----------
        * _shade (shade.Shade): The Honeybee-Shade to base the new component on.

    Returns:
    --------
        * (components.PhxComponentOpaque): A new PHX-Component for the HB-Shade.
    """

    new_compo = PhxComponentOpaque()

    new_compo.display_name = _shade.display_name
    new_compo.face_opacity = ComponentFaceOpacity.OPAQUE
    new_compo.exposure_exterior = ComponentExposureExterior.EXTERIOR
    new_compo.exposure_interior = -1
    new_compo.color_interior = ComponentColor.EXT_WALL_INNER
    new_compo.color_exterior = ComponentColor.EXT_WALL_INNER

    # -- Polygons
    phx_polygon = create_geometry.create_PhxPolygon_from_hb_Face(_shade)
    new_compo.add_polygons(phx_polygon)

    return new_compo


def add_hb_model_shades_to_variant(
    _var: project.PhxVariant,
    _hb_model: model.Model,
    _merge_faces: bool,
    _tolerance: float,
    _angle_tolerance_degrees: float,
) -> None:
    """ "Create shading PhxComponents from an HB-model's orphaned shades and add to the PhxVariant.

    This will group shade-faces by display-name and create a single component for each named-group.

    Arguments:
    ----------
        * _var (project.Variant): The PhxVariant to add the Shading Objects to.
        * _hb_model (model.Model): The Honeybee-Model to get the orphaned shades from.

    Returns:
    --------
        * None
    """

    # -- Group HB-Shades by their Display Name
    hb_shade_groups: defaultdict[str, List[Shade]] = defaultdict(list)
    for hb_shade in _hb_model.orphaned_shades:
        hb_shade_groups[hb_shade.display_name].append(hb_shade)

    # -- Create new component(s) from the groups
    grouped_shade_components: List[PhxComponentOpaque] = []
    for hb_shade_group in hb_shade_groups.values():
        # -- Merge HB-Shade-Faces
        if _merge_faces:
            face_groups = face_tools.group_hb_faces(hb_shade_group, _tolerance, _angle_tolerance_degrees)
            hb_shade_group: List[Shade] = []
            for face_group in face_groups:
                hb_shade_group += merge_hb_shades(face_group, _tolerance, _angle_tolerance_degrees)

        phx_compos = (create_new_component_from_orphaned_shade(s) for s in hb_shade_group)
        merged_phx_component: PhxComponentOpaque = reduce(operator.add, phx_compos)
        merged_phx_component.display_name = hb_shade_group[0].display_name

        grouped_shade_components.append(merged_phx_component)

    # -- Add the new component(s) to the Variant
    for shade_component in grouped_shade_components:
        _var.building.add_components(shade_component)

    return None



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Model class for the Ventilation worksheet various input items."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional

from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_data
from PHX.xl.xl_data import xl_writable


@dataclass
class VerificationInput:
    """Ventilation Worksheet input data item."""

    shape: shape_model.Verification
    input_type: str
    input_data: xl_writable
    input_unit: Optional[str] = None
    target_unit: Optional[str] = None

    @classmethod
    def item(
        cls,
        shape: shape_model.Verification,
        input_type: str,
        input_data: xl_writable,
        input_unit: Optional[str] = None,
        target_unit: Optional[str] = None,
    ) -> VerificationInput:
        """Create a new data-input item."""
        return cls(shape, input_type, input_data, input_unit, target_unit)

    @classmethod
    def enum(cls, shape: shape_model.Verification, input_type: str, input_enum_value: Enum) -> VerificationInput:
        """Create a new options-input item."""
        shape_data = getattr(shape, input_type).options
        return cls(shape, input_type, shape_data[str(input_enum_value.value)])

    def create_xl_item(self, _sheet_name: str, _row_num: int) -> xl_data.XlItem:
        """Returns a list of the XL Items to write for this Data item

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for

        Returns:
        --------
            * (XlItem): The XlItem to write to the sheet.
        """
        return xl_data.XlItem(
            sheet_name=_sheet_name,
            xl_range=f"{getattr(self.shape, self.input_type).input_column}{_row_num}",
            write_value=self.input_data,
            input_unit=self.input_unit,
            target_unit=self.target_unit,
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions used to convert a standard HBJSON Model over to WUFI Objects"""

import logging
from collections import defaultdict
from typing import List, Tuple, Union

from honeybee import model, room
from honeybee.aperture import Aperture
from honeybee_ph.properties.room import RoomPhProperties
from honeybee_ph.team import ProjectTeam, ProjectTeamMember

from PHX.from_HBJSON import cleanup, create_assemblies, create_schedules, create_shades, create_variant
from PHX.model.project import PhxProject, PhxProjectData, ProjectData_Agent

logger = logging.getLogger()


class MissingPropertiesError(Exception):
    def __init__(self, _lbt_obj):
        self.message = (
            f'Error: LBT Object "{_lbt_obj}" does not have a .properties attribute?\n'
            "Can not add the .ph to missing .properties attribute."
        )
        super().__init__(self.message)


def sort_hb_rooms_by_bldg_segment(_hb_rooms: Tuple[room.Room]) -> List[List[room.Room]]:
    """Returns Groups of Honeybee-Rooms broken up by properties.ph.ph_bldg_segment.identifier.

    Arguments:
    ----------
        * _hb_rooms (List[room.Room]): The list of Honeybee Rooms to sort into bins.

    Returns:
    --------
        * (List[List[room.Room]]): A list of the groups of Honeybee Rooms.
    """

    rooms_by_segment = defaultdict(list)
    for room in _hb_rooms:
        hb_room_prop_ph: RoomPhProperties = room.properties.ph  # type: ignore
        rooms_by_segment[hb_room_prop_ph.ph_bldg_segment.identifier].append(room)
    return list(rooms_by_segment.values())


def hb_team_member_to_phx_agent(
    _hb_team_member: ProjectTeamMember,
) -> ProjectData_Agent:
    """Return a new PHX ProjectData_Agent with data based on a HB ProjectTeamMember."""
    return ProjectData_Agent(
        _hb_team_member.name,
        _hb_team_member.street,
        _hb_team_member.city,
        _hb_team_member.post_code,
        _hb_team_member.telephone,
        _hb_team_member.email,
    )


def get_project_data_from_hb_model(_hb_model: model.Model) -> PhxProjectData:
    """Return a new PhxProjectData with all team-member info based on an HB Model."""
    hb_proj_team = _hb_model.properties.ph.team  # type: ProjectTeam # type: ignore
    new_project_data = PhxProjectData()
    new_project_data.customer = hb_team_member_to_phx_agent(hb_proj_team.customer)
    new_project_data.owner = hb_team_member_to_phx_agent(hb_proj_team.owner)
    new_project_data.building = hb_team_member_to_phx_agent(hb_proj_team.building)
    new_project_data.designer = hb_team_member_to_phx_agent(hb_proj_team.designer)
    return new_project_data


def get_hb_apertures(_hb_model: model.Model) -> List[Aperture]:
    """Return a list of all the HB Apertures in the Model."""
    return [aperture for room in _hb_model.rooms for face in room.faces for aperture in face.apertures]


def convert_hb_model_to_PhxProject(
    _hb_model: model.Model,
    _group_components: bool = True,
    _merge_faces: Union[bool, float] = False,
    _merge_spaces_by_erv: bool = False,
) -> PhxProject:
    """Return a complete WUFI Project object with values based on the HB Model

    Arguments:
    ----------
        * _hb_model (model.Model): The Honeybee Model to base the WUFI Project on

        * _group_components (bool): default=True. Set to true to have the converter
            group the components by assembly-type.

        * _merge_faces (bool | float): default=False. Set to true to have the converter try and
            group together co-planar faces in the output room using the HB model tolerance.
            If a number is given, it will be used as the tolerance for merging faces.

        * _merge_spaces_by_erv (bool): default=False. Set to true to have the converter
            merge all the spaces by ERV zones. This is sometimes required by Phius for
            large buildings with multiple ERV zones.

    Returns:
    --------
        * (PhxProject): The new WUFI Project object.
    """

    phx_project = PhxProject()
    phx_project.project_data = get_project_data_from_hb_model(_hb_model)
    create_assemblies.build_opaque_assemblies_from_HB_model(phx_project, _hb_model)
    create_assemblies.build_transparent_assembly_types_from_HB_Model(phx_project, get_hb_apertures(_hb_model))
    create_schedules.add_all_HB_schedules_to_PHX_Project(phx_project, _hb_model)

    # -- TODO: Make all these operations if..else... with flags in the func arguments.

    # -- Merge the rooms together by their Building Segment, Add to the Project
    # -- then create a new variant from the merged room.
    # -- try and weld the vertices too in order to reduce load-time.
    for room_group in sort_hb_rooms_by_bldg_segment(_hb_model.rooms):  # type: ignore
        # -- Configure the merge_faces and merge_face_tolerance
        if isinstance(_merge_faces, bool):
            merge_faces: bool = _merge_faces
            merge_face_tolerance: float = _hb_model.tolerance
        else:
            merge_faces: bool = True
            merge_face_tolerance: float = _merge_faces

        merged_hb_room = cleanup.merge_rooms(room_group, merge_face_tolerance, _hb_model.angle_tolerance, merge_faces)

        new_variant = create_variant.from_hb_room(
            _hb_room=merged_hb_room,
            _assembly_dict=phx_project.assembly_types,
            _window_type_dict=phx_project.window_types,
            _vent_sched_collection=phx_project.utilization_patterns_ventilation,
            _occ_sched_collection=phx_project.utilization_patterns_occupancy,
            _lighting_sched_collection=phx_project.utilization_patterns_lighting,
            _group_components=_group_components,
            _merge_spaces_by_erv=_merge_spaces_by_erv,
            _tolerance=_hb_model.tolerance,
        )

        new_variant = cleanup.weld_vertices(new_variant)

        create_shades.add_hb_model_shades_to_variant(
            new_variant,
            _hb_model,
            _merge_faces=merge_faces,
            _tolerance=_hb_model.tolerance,
            _angle_tolerance_degrees=_hb_model.angle_tolerance,
        )

        phx_project.add_new_variant(new_variant)

    return phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Program Class for organizing Lighting Data."""

from __future__ import annotations

from dataclasses import dataclass, field

from PHX.model.loads.lighting import PhxLoadLighting
from PHX.model.schedules.lighting import PhxScheduleLighting


@dataclass
class PhxProgramLighting:
    """A PHX Program for the Lighting with a load and schedule."""

    display_name: str = "Unnamed_Lighting_Program"
    load: PhxLoadLighting = field(default_factory=PhxLoadLighting)
    schedule: PhxScheduleLighting = field(default_factory=PhxScheduleLighting)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Ventilation Load."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, ClassVar


@dataclass
class PhxLoadVentilation:
    """A PHX Load for the Ventilation."""

    flow_supply: float = 0.0
    flow_extract: float = 0.0
    flow_transfer: float = 0.0

    @property
    def total_airflow(self) -> float:
        """Returns the total airflow for the ventilation load."""
        return self.flow_supply + self.flow_extract + self.flow_transfer

    def __add__(self, other: PhxLoadVentilation) -> PhxLoadVentilation:
        """Combine two PhxLoadVentilation objects."""
        new_load = PhxLoadVentilation()
        new_load.flow_supply = self.flow_supply + other.flow_supply
        new_load.flow_extract = self.flow_extract + other.flow_extract
        new_load.flow_transfer = self.flow_transfer + other.flow_transfer
        return new_load



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Valid 'types' for Building Elements."""

from __future__ import annotations

from enum import Enum


class ComponentFaceType(Enum):
    NONE = 0
    WALL = 1
    FLOOR = 2
    ROOF_CEILING = 3
    AIR_BOUNDARY = 3
    WINDOW = 4
    ADIABATIC = 5
    CUSTOM = 6

    @classmethod
    def by_angle(cls, _angle: float) -> ComponentFaceType:
        if _angle < 70.0:
            return cls.ROOF_CEILING
        if _angle > 110.0:
            return cls.FLOOR
        return cls.WALL


class ComponentExposureExterior(Enum):
    NONE = 0
    EXTERIOR = -1
    GROUND = -2
    SURFACE = -3


class ComponentFaceOpacity(Enum):
    OPAQUE = 1
    TRANSPARENT = 2
    AIRBOUNDARY = 3


class ComponentColor(Enum):
    EXT_WALL_INNER = 1
    EXT_WALL_OUTER = 2
    INNER_WALL = 3
    WINDOW = 4
    FLOOR = 5
    CEILING = 6
    SLOPED_ROOF_OUTER = 7
    SLOPED_ROOF_INNER = 8
    SLOPED_ROOF_THATCH = 9  # WTF?
    FLAT_ROOF_OUTER = 10
    FLAT_ROOF_INNER = 11
    SURFACE_GROUND_CONTACT = 12
    GROUND_ABOVE = 13
    GROUND_BENEATH = 14


class ThermalBridgeType(Enum):
    AMBIENT = 15
    PERIMETER = 16
    UNDERGROUND = 17


class SpecificHeatCapacity(Enum):
    LIGHTWEIGHT = 1
    MIXED = 2
    MASSIVE = 3


class WufiVolumeGrossMode(Enum):
    USER_DEFINED = 6
    FROM_VISUALIZED_COMPONENTS = 7


class WufiVolumeNetMode(Enum):
    ESTIMATED_FROM_GROSS = 4
    FROM_GROSS_AND_COMPONETS = 5
    USER_DEFINED = 6
    FROM_VISUALIZED_COMPONENTS = 7


class WufiWeightedFloorAreaMode(Enum):
    FROM_VISUALIZED_GEOMETRY = 2
    ESTIMATED_FROM_GROSS_VOLUME = 4
    USER_DEFINED = 6


class ZoneType(Enum):
    SIMULATED = 1
    ATTACHED = 2


class AttachedZoneType(Enum):
    NONE = 0
    UNHEATED_SPACE = 1
    UNHEATED_CELLAR = 2
    UNHEATED_CRAWLSPACE = 3
    UNHEATED_WINTER_GARDEN = 4
    UNHEATED_ATTIC = 5
    CONDITIONED = 6


class WindExposureType(Enum):
    SEVERAL_SIDES_EXPOSED_NO_SCREENING = 1
    SEVERAL_SIDES_EXPOSED_MODERATE_SCREENING = 2
    SEVERAL_SIDES_EXPOSED_HIGH_SCREENING = 3
    ONE_SIDE_EXPOSED_NO_SCREENING = 4
    ONE_SIDE_EXPOSED_MODERATE_SCREENING = 5
    ONE_SIDE_EXPOSED_HIGH_SCREENING = 7
    USER_DEFINED = 6



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP 'Electricity' worksheet."""

from __future__ import annotations

from typing import List, Type

from PHX.model import elec_equip
from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import electricity_item
from PHX.PHPP.phpp_model.electricity_item import PHPPReadAddress
from PHX.xl import xl_app, xl_data


class Electricity:
    """IO Controller for PHPP "Electricity" worksheet."""

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Electricity) -> None:
        self.xl = _xl
        self.shape = shape
        self.device_map = elec_equip.get_device_type_map()

    def _turn_off_all_equipment(self) -> None:
        """Sets all the 'used' values to 0 to reset the sheet before writing new equipment."""
        for item in self.shape.input_rows:
            # Some items cannot be turned off....
            excluded = [
                "clothes_drying",
                "cooking",
                "consumer_elec",
                "lighting",
                "small_appliances",
            ]
            if item[0] in excluded:
                continue

            self.xl.write_xl_item(xl_data.XlItem(self.shape.name, f"{self.shape.input_columns.used}{item[1].data}", 0))

    def write_equipment(self, _equipment_inputs: List[electricity_item.ElectricityItemXLWriter]) -> None:
        """Write a list of equipment-input objects to the Worksheet."""
        self._turn_off_all_equipment()

        for equip_input in _equipment_inputs:
            for item in equip_input.create_xl_items(self.shape):
                self.xl.write_xl_item(item)

    def build_phx_device_from_phpp(self, _reader: electricity_item.ReaderDataItem) -> elec_equip.PhxElectricalDevice:
        """Build a PHX Electrical Device object from the PHPP worksheet data."""

        # -- Get the right device class based on the device-type
        cls = self.device_map[_reader.type]
        phx_elec_device = cls()

        # -- Build the new Device using the input data from the PHPP Reader
        for phpp_read_address in _reader.data:
            phpp_data = self.xl.get_single_data_item(self.shape.name, phpp_read_address.phpp_address)
            setattr(phx_elec_device, phpp_read_address.attr_name, phpp_data)
        return phx_elec_device

    def get_phx_elec_devices(self) -> List[elec_equip.PhxElectricalDevice]:
        """Read the Device data from the PHPP worksheet and return a list of PhxElectricalDevice objects."""
        # -- Setup the reader class
        reader = electricity_item.ElectricityItemXLReader(self.shape)

        phx_elec_devices = [
            self.build_phx_device_from_phpp(reader._dishwasher),
            self.build_phx_device_from_phpp(reader._clothes_washer),
            self.build_phx_device_from_phpp(reader._clothes_dryer),
            self.build_phx_device_from_phpp(reader._refrigerator),
            self.build_phx_device_from_phpp(reader._fridge_freezer),
            self.build_phx_device_from_phpp(reader._freezer),
            self.build_phx_device_from_phpp(reader._cooktop),
            self.build_phx_device_from_phpp(reader._mel),
            self.build_phx_device_from_phpp(reader._lighting_interior),
            self.build_phx_device_from_phpp(reader._lighting_exterior),
        ]

        return phx_elec_devices



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Ventilation Ducting Distribution Objects."""


from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional

from ladybug_geometry.geometry3d.polyline import LineSegment3D

from PHX.model.enums.hvac import PhxVentDuctType


@dataclass
class PhxDuctSegment:
    """An individual Duct Segment Segment."""

    identifier: str
    display_name: str
    geometry: LineSegment3D
    diameter_m: float
    height_m: Optional[float]
    width_m: Optional[float]
    insulation_thickness_m: float
    insulation_conductivity_wmk: float
    insulation_reflective: bool

    @property
    def length(self) -> float:
        return self.geometry.length

    @property
    def diameter_mm(self) -> float:
        """Return the diameter in MM."""
        return self.diameter_m * 1000.0

    @property
    def height_mm(self) -> Optional[float]:
        """Return the height in MM."""
        if self.height_m:
            return self.height_m * 1000.0
        else:
            return None

    @property
    def width_mm(self) -> Optional[float]:
        """Return the width in MM."""
        if self.width_m:
            return self.width_m * 1000.0
        else:
            return None

    @property
    def insulation_thickness_mm(self) -> float:
        """Return the insulation-thickness in MM."""
        return self.insulation_thickness_m * 1000.0


@dataclass
class PhxDuctElement:
    """A Duct Element / Run made of one or more PhxDuctSegments."""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)

    identifier: str
    display_name: str
    duct_type: PhxVentDuctType = field(init=False, default=PhxVentDuctType.SUPPLY)
    vent_unit_id: int
    _segments: Dict = field(default_factory=dict)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def quantity(self) -> int:
        return 1

    @property
    def segments(self) -> List[PhxDuctSegment]:
        return list(self._segments.values())

    @property
    def length_m(self) -> float:
        return sum(_.length for _ in self.segments)

    @property
    def diameter_mm(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * seg.diameter_mm

        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def height_mm(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * (seg.height_mm or 0.0)

        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def width_mm(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * (seg.width_mm or 0.0)

        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def insulation_thickness_mm(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * seg.insulation_thickness_mm

        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def insulation_conductivity_wmk(self) -> float:
        weighted_total = 0.0
        for seg in self.segments:
            weighted_total += seg.length * seg.insulation_conductivity_wmk

        try:
            return weighted_total / self.length_m
        except ZeroDivisionError:
            return 0.0

    @property
    def duct_shape(self) -> int:
        if self.height_mm and self.width_mm:
            return 2  # Rectangular Duct
        else:
            return 1  # Round Duct

    @property
    def is_reflective(self) -> bool:
        return any({seg.insulation_reflective for seg in self.segments})

    @property
    def assigned_vent_unit_ids(self) -> List[int]:
        return [self.vent_unit_id]

    def add_segment(self, _s: PhxDuctSegment) -> None:
        self._segments[_s.identifier] = _s



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Run script to convert an HBJSON file over to WUFI XML format."""

import glob
import logging
import os
import pathlib
import sys
from datetime import datetime
from typing import List, Tuple, Union

from PHX.from_HBJSON import create_project, read_HBJSON_file
from PHX.to_WUFI_XML import _bug_fixes, xml_builder, xml_txt_to_file


class InputFileError(Exception):
    def __init__(self, path) -> None:
        self.msg = f"\nError: Cannot find HBJSON file: {path}"
        super().__init__(self.msg)


def resolve_paths(_args: List[str]) -> Tuple[pathlib.Path, pathlib.Path]:
    """Sort out the file input and output paths. Make the output directory if needed.

    Arguments:
    ----------
        * _args (Tuple): sys.args Tuple of inputs.

    Returns:
    --------
        * Tuple
            - [1] (str): The HBJSON Source file path.
            - [2] (str): The WUFI XML Target file name.
            - [3] (str): The WUFI XML Target directory path.
    """

    print("> Resolving file paths...")
    src = pathlib.Path(_args[1])
    if not src.exists():
        raise InputFileError(src)

    target_file = f"{_args[2]}.xml"
    target_dir = pathlib.Path(_args[3])

    if not target_dir.exists():
        os.mkdir(target_dir)

    target = pathlib.Path(target_dir, target_file)

    return src, target


def group_components(_args: List[str]) -> bool:
    """Return the 'group_components' boolean from the sys.args Tuple.

    Arguments:
    ----------
        * _args (Tuple): sys.args Tuple of inputs.
            - [4] (str): "True" or "False" string.

    Returns:
    --------
        * bool: True if the user wants to group components.
    """
    return _args[4].lower() == "true"


def merge_faces(_args: List[str]) -> Union[bool, float]:
    """Return the 'merge_faces' as bool | float from the sys.args Tuple.

    Arguments:
    ----------
        * _args (Tuple): sys.args Tuple of inputs.
            - [5] (str): "True", "False", or float value as a string.

    Returns:
    --------
        * bool | float: True, False, or a float value for the tolerance.
    """
    if _args[5].lower() == "true":
        return True
    elif _args[5].lower() == "false":
        return False
    else:
        return float(_args[5])


def merge_spaces_by_erv(_args: List[str]) -> bool:
    """Return the 'merge_spaces_by_erv' boolean from the sys.args Tuple.

    Arguments:
    ----------
        * _args (Tuple): sys.args Tuple of inputs.
            - [6] (str): "True" or "False" string.

    Returns:
    --------
        * bool: True if the user wants to group components.
    """
    return _args[6].lower() == "true"


def log_level(_args: List[str]) -> int:
    """Return the log_level from the sys.args Tuple.

    Arguments:
    ----------
        * _args (Tuple): sys.args Tuple of inputs.
            - [6] (str): "0", "10", "22", "30" .....

    Returns:
    --------
        * int: The logging level.
    """
    try:
        return int(_args[7])
    except Exception:
        return 0


def remove_old_logs(directory: pathlib.Path, max_files: int = 10):
    """Remove old log files from the directory.

    Arguments:
    ----------
        * directory (pathlib.Path): The directory to remove the old logs from.
        * max_files (int): The maximum number of log files to keep.

    Returns:
    --------
        * None
    """

    all_log_files = glob.glob(os.path.join(directory, "*.log"))
    if len(all_log_files) > max_files:
        sorted_files = sorted(all_log_files, key=os.path.getmtime)
        oldest_log_file = sorted_files[0]
        os.remove(oldest_log_file)


def setup_logging_dir(_source_file_path: pathlib.Path) -> pathlib.Path:
    """Return logging directory. Create it if needed.

    Arguments:
    ----------
        * _source_file_path (pathlib.Path): The base directory to create the logging directory in.

    Returns:
    --------
        * pathlib.Path: The logging directory.
    """
    log_dir = _source_file_path.parent / "PHX_Logs"

    if not log_dir.exists():
        os.mkdir(log_dir)

    return log_dir


def startup_logging(_log_level: int) -> logging.Logger:
    """Setup the logging. Create a new dir if needed..

    Arguments:
    ----------
        * _log_level (int): The logging level.

    Returns:
    --------
        * logging.Logger: The root logger.
    """

    logger = logging.getLogger()  # Root Logger
    logger.setLevel(_log_level)

    # -- Set Format
    current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(funcName)s - %(message)s")

    # -- Setup the STDERR log stream-handler for stderr
    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setLevel(logging.WARNING)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    # -- Setup the STDOUT log stream-handler for stdout
    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setLevel(logging.INFO)
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)

    if _log_level > 0:
        # -- Find the right path, create if needed. Clean up old logs.
        log_path = setup_logging_dir(SOURCE_FILE)
        remove_old_logs(log_path, 10)

        # -- Setup the log file-handle
        file_handler = logging.FileHandler(log_path / f"PHX_{current_time}.log")
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

        logger.info(f"> LOGGING TO: {log_path / 'PHX.log'}")

    return logger


if __name__ == "__main__":
    print("- " * 50)

    # --- Input / Output file Path
    # -------------------------------------------------------------------------
    SOURCE_FILE, TARGET_FILE_XML = resolve_paths(sys.argv)
    GROUP_COMPONENTS = group_components(sys.argv)
    MERGE_FACES = merge_faces(sys.argv)
    MERGE_SPACES_BY_ERV = merge_spaces_by_erv(sys.argv)
    LOG_LEVEL = log_level(sys.argv)

    ## -- Setup the logging
    logger = startup_logging(LOG_LEVEL)
    logger.info(f"Logging with log-level: {LOG_LEVEL}")
    logger.info(f"Group Components: {GROUP_COMPONENTS}")
    logger.info(f"Merging Faces: {MERGE_FACES}")
    logger.info(f"Merging Spaces by ERV: {MERGE_SPACES_BY_ERV}")

    # --- Read in the existing HB_JSON and re-build the HB Objects
    # -------------------------------------------------------------------------
    logger.info(f"> Reading in the HBJSON file: ./{SOURCE_FILE}")
    hb_json_dict = read_HBJSON_file.read_hb_json_from_file(SOURCE_FILE)
    hb_model = read_HBJSON_file.convert_hbjson_dict_to_hb_model(hb_json_dict)

    # --- Generate the WUFI Project file.
    logger.info(f'> Generating the PHX-Project from the Honeybee-Model: "{hb_model}"')
    phx_Project = create_project.convert_hb_model_to_PhxProject(
        hb_model,
        _group_components=GROUP_COMPONENTS,
        _merge_faces=MERGE_FACES,
        _merge_spaces_by_erv=MERGE_SPACES_BY_ERV,
    )

    # --- Apply the WUFI-Passive Cooling Bug fix (200 KW limit)
    phx_Project = _bug_fixes.split_cooling_into_multiple_systems(phx_Project)

    # --- Output the WUFI Project as an XML Text File
    # -------------------------------------------------------------------------
    logger.info(f'> Generating XML Text for the PHX-Project: "{phx_Project}"')
    xml_txt = xml_builder.generate_WUFI_XML_from_object(phx_Project)

    logger.info(f"> Saving the XML file to: ./{TARGET_FILE_XML}")
    xml_txt_to_file.write_XML_text_file(TARGET_FILE_XML, xml_txt)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP Climate worksheet."""

from __future__ import annotations

from dataclasses import dataclass
from typing import List

from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import verification_data
from PHX.xl import xl_app


class VerificationInputLocation:
    """Generic input item for Verification worksheet items."""

    def __init__(
        self,
        _xl: xl_app.XLConnection,
        _sheet_name: str,
        _search_col: str,
        _search_item: str,
        _input_row_offset: int,
    ):
        self.xl = _xl
        self.sheet_name = _sheet_name
        self.search_col = _search_col
        self.search_item = _search_item
        self.input_row_offset = _input_row_offset

    def find_input_row(self, _row_start: int = 1, _row_end: int = 200) -> int:
        """Return the row number where the search-item is found input."""
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.sheet_name,
            _col=self.search_col,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data, start=_row_start):
            if self.search_item in str(val):
                return i + self.input_row_offset

        raise Exception(
            f'\n\tError: Not able to find the "{self.search_item}" input '
            f'section of the "{self.sheet_name}" worksheet? Please be sure '
            f'the item is note with the "{self.search_item}" flag in column {self.search_col}?'
        )


@dataclass
class TeamMemberData:
    """A Dataclass to store team-member information when read in from the PHPP."""

    name: str
    street_name: str
    post_code: str
    city: str
    state: str
    country: str

    @classmethod
    def from_raw_excel_data(cls, _xl_data) -> TeamMemberData:
        """Create a new TeamMemberData object from raw excel data read in from PHPP

        Arguments:
        ----------
            * _xl_data: List[List[str]]: A list of lists containing the data read
                in from PHPP.
        Returns:
        --------
            * (TeamMemberData): A new TeamMemberData object with the values from the
                input data set.
        """
        try:
            return cls(
                name=_xl_data[0][0],
                street_name=_xl_data[1][0],
                post_code=_xl_data[2][0],
                city=_xl_data[2][1],
                state=_xl_data[3][0],
                country=_xl_data[3][2],
            )
        except:
            msg = f"Error reading in Team Member Data from the PHPP? Got:\n{_xl_data}"
            raise Exception(msg)


class Verification:
    def __init__(self, _xl: xl_app.XLConnection, _shape: shape_model.Verification):
        self.xl = _xl
        self.shape = _shape

    def _create_input_location_object(
        self, _phpp_model_obj: verification_data.VerificationInput
    ) -> VerificationInputLocation:
        """Create and setup the VerificationInputLocation object with the correct data."""
        phpp_obj_shape: shape_model.VerificationInputItem = getattr(self.shape, _phpp_model_obj.input_type)
        return VerificationInputLocation(
            _xl=self.xl,
            _sheet_name=self.shape.name,
            _search_col=phpp_obj_shape.locator_col,
            _search_item=phpp_obj_shape.locator_string,
            _input_row_offset=phpp_obj_shape.input_row_offset,
        )

    def write_item(self, _phpp_model_obj: verification_data.VerificationInput) -> None:
        """Write the VerificationInputItem item out to the PHPP Verification Worksheet."""
        input_object = self._create_input_location_object(_phpp_model_obj)
        input_row = input_object.find_input_row()
        xl_item = _phpp_model_obj.create_xl_item(self.shape.name, input_row)
        self.xl.write_xl_item(xl_item)

    def read_architect(self) -> TeamMemberData:
        """Return a TeamMemberData object with the architect info from PHPP."""
        data = self.xl.get_data(self.shape.name, "F18:H21")
        return TeamMemberData.from_raw_excel_data(data)

    def read_energy_consultant(self) -> TeamMemberData:
        """Return a TeamMemberData object with the consultant info from PHPP."""
        data = self.xl.get_data(self.shape.name, "F23:H26")
        return TeamMemberData.from_raw_excel_data(data)

    def read_building(self) -> TeamMemberData:
        """Return a TeamMemberData object with the Building address from PHPP."""
        data = self.xl.get_data(self.shape.name, "K5:M8")
        return TeamMemberData.from_raw_excel_data(data)

    def read_site_owner(self) -> TeamMemberData:
        """Return a TeamMemberData object with the owner info from PHPP."""
        data = self.xl.get_data(self.shape.name, "K13:M16")
        return TeamMemberData.from_raw_excel_data(data)

    def read_mech_engineer(self) -> TeamMemberData:
        data = self.xl.get_data(self.shape.name, "K18:M21")
        """Return a TeamMemberData object with the ME info from PHPP."""
        return TeamMemberData.from_raw_excel_data(data)

    def read_ph_certification(self) -> TeamMemberData:
        """Return a TeamMemberData object with the Certifier info from PHPP."""
        data = self.xl.get_data(self.shape.name, "K23:M26")
        return TeamMemberData.from_raw_excel_data(data)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Model class for a PHPP Shading/Shading-Entry row"""

from dataclasses import dataclass
from functools import partial
from typing import List, Tuple

from PHX.model import components
from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_data


@dataclass
class ShadingRow:
    """Model class for a single Window's Shading entry row."""

    __slots__ = (
        "shape",
        "shading_dims",
        "winter_shading_factor",
        "summer_shading_factor",
    )
    shape: shape_model.Shading
    shading_dims: components.PhxApertureShadingDimensions
    winter_shading_factor: float
    summer_shading_factor: float

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        col = getattr(self.shape.shading_rows.inputs, _field_name).column
        return f"{col}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.shading_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Surface Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemShading = partial(xl_data.XlItem, _sheet_name)
        items: List[xl_data.XlItem] = [
            XLItemShading(
                create_range("h_hori"),
                self.shading_dims.h_hori,
                "M",
                self._get_target_unit("h_hori"),
            ),
            XLItemShading(
                create_range("d_hori"),
                self.shading_dims.d_hori,
                "M",
                self._get_target_unit("d_hori"),
            ),
            XLItemShading(
                create_range("o_reveal"),
                self.shading_dims.o_reveal,
                "M",
                self._get_target_unit("o_reveal"),
            ),
            XLItemShading(
                create_range("d_reveal"),
                self.shading_dims.d_reveal,
                "M",
                self._get_target_unit("d_reveal"),
            ),
            XLItemShading(
                create_range("o_over"),
                self.shading_dims.o_over,
                "M",
                self._get_target_unit("o_over"),
            ),
            XLItemShading(
                create_range("d_over"),
                self.shading_dims.d_over,
                "M",
                self._get_target_unit("d_over"),
            ),
            XLItemShading(create_range("r_other_winter"), self.winter_shading_factor),
            XLItemShading(create_range("r_other_summer"), self.summer_shading_factor),
        ]

        items_merged = xl_data.merge_xl_item_row(items)

        return items_merged



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Valid 'types' for Mech Equipment Options."""

from enum import Enum

from honeybee_phhvac.hot_water_devices import PhHvacHotWaterTankType


class PhxFuelType(Enum):
    NATURAL_GAS = 1
    OIL = 2
    WOOD_LOG = 3
    WOOD_PELLET = 4
    # HARD_COAL_CGS_70_PHC = 'HARD_COAL_CGS_70_PHC'
    # HARD_COAL_CGS_35_PHC = 'HARD_COAL_CGS_35_PHC'
    # HARD_COAL_HS_0_PHC = 'HARD_COAL_HS_0_PHC'
    # GAS_CGS_70_PHC = 'GAS_CGS_70_PHC'
    # GAS_CGS_35_PHC = 'GAS_CGS_35_PHC'
    # GAS_HS_0_PHC = 'GAS_HS_0_PHC'
    # OIL_CGS_70_PHC = 'OIL_CGS_70_PHC'
    # OIL_CGS_35_PHC = 'OIL_CGS_35_PHC'
    # OIL_HS_0_PHC = 'OIL_HS_0_PHC'


class SystemType(Enum):
    ANY = 0
    VENTILATION = 1
    ELECTRIC = 2
    BOILER = 3
    DISTRICT_HEAT = 4
    HEAT_PUMP = 5
    USER_DEFINED = 7
    WATER_STORAGE = 8
    PHOTOVOLTAIC = 10


class DeviceType(Enum):
    VENTILATION = 1
    ELECTRIC = 2
    BOILER = 3
    DISTRICT_HEAT = 4
    HEAT_PUMP = 5
    WATER_STORAGE = 8
    PHOTOVOLTAIC = 10


class HeatPumpType(Enum):
    COMBINED = 2
    ANNUAL = 3
    RATED_MONTHLY = 4
    HOT_WATER = 5


class CoolingType(Enum):
    NONE = 0
    VENTILATION = 1
    RECIRCULATION = 2
    DEHUMIDIFICATION = 3
    PANEL = 4


class PhxHotWaterPipingCalcMethod(Enum):
    SIMPLIFIED_INDIVIDUAL_PIPES = 1
    SIMPLIFIED_HOT_WATER_CALCULATOR = 2
    HOT_WATER_PIPING_UNIT_METHOD = 3
    HOT_WATER_PIPING_FLOOR_METHOD = 4


class PhxHotWaterPipingMaterial(Enum):
    COPPER_M = 1
    COPPER_L = 2
    COPPER_K = 3
    CPVC_CTS_SDR = 4
    CPVC_SCH_40 = 5
    PEX = 6
    PE = 7
    PEX_CTS_SDR = 8


class PhxHotWaterPipingInchDiameterType(Enum):
    _0_3_8_IN = 1
    _0_1_2_IN = 2
    _0_5_8_IN = 3
    _0_3_4_IN = 4
    _1_0_0_IN = 5
    _1_1_4_IN = 6
    _1_1_2_IN = 7
    _2_0_0_IN = 8

    @property
    def name_as_float(self) -> float:
        """Return the name as a float. Oh my fucking god IP units are so stupid."""
        # -- Remove the '_IN' suffix
        name = str(self.name).split("_IN")[0]

        # -- Remove the prefix underscore
        name = name[1:]

        # -- Split the fraction
        parts = str(name).split("_")
        try:
            fraction = "{}".format(float(parts[1]) / float(parts[2]))
        except ZeroDivisionError:
            fraction = 0.0

        # -- Return the whole
        return float(parts[0]) + float(fraction)

    @classmethod
    def nearest_key(cls, value: float) -> "PhxHotWaterPipingInchDiameterType":
        """Return a new Enum object which most closely matches the supplied value."""
        return min(PhxHotWaterPipingInchDiameterType, key=lambda d: abs(d.name_as_float - value))


class PhxHotWaterInputOptions(Enum):
    SPEC_TOTAL_LOSSES = 1
    SPEC_STANDBY_LOSSES = 2
    TOTAL_LOSSES = 3


class PhxHotWaterTankType(Enum):
    NONE = 0
    DHW_AND_HEATING = 1
    DHW_ONLY = 2

    @classmethod
    def from_hbph_type(cls, _hbph_type):
        """Return a new PhxHotWaterTankType based on the Honeybee-PH Tank Type."""
        _mapping = {
            "0-NO STORAGE TANK": cls.NONE,
            "1-DHW AND HEATING": cls.DHW_AND_HEATING,
            "2-DHW ONLY": cls.DHW_ONLY,
        }
        return _mapping[_hbph_type]


class PhxHotWaterSelectionUnitsOrFloors(Enum):
    PH_CASE = 1
    USER_DETERMINED = 2


class PhxExhaustVentType(Enum):
    DRYER = 1
    KITCHEN_HOOD = 2
    USER_DEFINED = 3


class PhxVentDuctType(Enum):
    SUPPLY = 1
    EXHAUST = 2


class PhxSupportiveDeviceType(Enum):
    HEAT_CIRCULATING_PUMP = 4
    DHW_CIRCULATING_PUMP = 6
    DHW_STORAGE_LOAD_PUMP = 7
    OTHER = 10


class PhxSummerBypassMode(Enum):
    NONE = 1
    TEMP_CONTROLLED = 2
    ENTHALPY_CONTROLLED = 3
    ALWAYS = 4



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Model class for a PHPP Electricity / Equipment row input."""

from dataclasses import dataclass
from typing import Any, Dict, List, NamedTuple, Tuple

from PHX.model import elec_equip
from PHX.model.enums.elec_equip import ElectricEquipmentType
from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_data

# -----------------------------------------------------------------------------
# -- Worksheet Writer


@dataclass
class ElectricityItemXLWriter:
    """Model class for a single Electric-Equipment item entry row."""

    __slots__ = "phx_equipment"
    phx_equipment: Any

    def create_xl_items(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        """Returns a list of xl_data.XlItem or raises and Error if equipment type is unrecognized."""
        xl_item_functions = {
            "PhxDeviceDishwasher": self._dishwasher,
            "PhxDeviceClothesWasher": self._clothes_washer,
            "PhxDeviceClothesDryer": self._clothes_dryer,
            "PhxDeviceRefrigerator": self._refrigerator,
            "PhxDeviceFreezer": self._freezer,
            "PhxDeviceFridgeFreezer": self._fridge_freezer,
            "PhxDeviceCooktop": self._cooktop,
            "PhxDeviceMEL": self._mel,
            "PhxDeviceLightingInterior": self._lighting_interior,
            "PhxDeviceLightingExterior": self._lighting_exterior,
            "PhxDeviceLightingGarage": self._lighting_garage,
            "PhxDeviceCustomElec": self._custom_elec,
            "PhxDeviceCustomLighting": self._custom_lighting,
            "PhxDeviceCustomMEL": self._custom_mel,
        }
        try:
            return xl_item_functions[self.phx_equipment.__class__.__name__](_shape)
        except KeyError:
            raise NotImplementedError(
                f"No matching XL-write function found for equipment type: '{self.phx_equipment.__class__.__name__}'"
            )

    def _dishwasher(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceDishwasher = self.phx_equipment
        cols = _shape.input_columns
        rows = _shape.input_rows
        address_used = f"{cols.used}{rows.dishwasher.data}"
        address_inside = f"{cols.in_conditioned_space}{rows.dishwasher.data}"
        address_demand = f"{cols.energy_demand_per_use}{rows.dishwasher.data}"
        address_selection = f"{cols.selection}{rows.dishwasher.selection}"

        items: List[Tuple[str, xl_data.xl_writable]] = [
            (address_used, 1),
            (address_inside, str(int(equip.in_conditioned_space or 1))),
            (address_demand, equip.energy_demand_per_use),
            (
                address_selection,
                rows.dishwasher.selection_options[str(equip.water_connection)],
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _clothes_washer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceClothesWasher = self.phx_equipment
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (f"{_shape.input_columns.used}{_shape.input_rows.clothes_washing.data}", 1),
            (
                f"{_shape.input_columns.in_conditioned_space}{_shape.input_rows.clothes_washing.data}",
                str(int(equip.in_conditioned_space or 1)),
            ),
            (
                f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.clothes_washing.data}",
                equip.energy_demand_per_use,
            ),
            (
                f"{_shape.input_columns.selection}{_shape.input_rows.clothes_washing.selection}",
                _shape.input_rows.clothes_washing.selection_options[str(equip.water_connection)],
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _clothes_dryer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceClothesDryer = self.phx_equipment
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f"{_shape.input_columns.in_conditioned_space}{_shape.input_rows.clothes_drying.data}",
                str(int(equip.in_conditioned_space or 1)),
            ),
            (
                f"{_shape.input_columns.selection}{_shape.input_rows.clothes_drying.selection}",
                _shape.input_rows.clothes_drying.selection_options[str(equip.dryer_type)],
            ),
        ]

        # -- Add energy consumption, location depends on fuel type
        if equip.dryer_type == 6:
            # Gas dryer
            items.append(
                (
                    f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.clothes_drying.selection}",
                    equip.gas_consumption,
                )
            )
        else:
            # non-Gas dryer
            items.append(
                (
                    f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.clothes_drying.data}",
                    equip.energy_demand_per_use,
                ),
            )

        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _refrigerator(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceRefrigerator = self.phx_equipment
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (f"{_shape.input_columns.used}{_shape.input_rows.refrigerator.data}", 1),
            (
                f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.refrigerator.data}",
                equip.energy_demand_per_use,
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _fridge_freezer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceFridgeFreezer = self.phx_equipment
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (f"{_shape.input_columns.used}{_shape.input_rows.fridge_freezer.data}", 1),
            (
                f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.fridge_freezer.data}",
                equip.energy_demand_per_use,
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _freezer(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceFreezer = self.phx_equipment
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (f"{_shape.input_columns.used}{_shape.input_rows.freezer.data}", 1),
            (
                f"{_shape.input_columns.in_conditioned_space}{_shape.input_rows.freezer.data}",
                str(int(equip.in_conditioned_space or 1)),
            ),
            (
                f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.freezer.data}",
                equip.energy_demand_per_use,
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _cooktop(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        equip: elec_equip.PhxDeviceCooktop = self.phx_equipment
        items: List[Tuple[str, xl_data.xl_writable]] = [
            (
                f"{_shape.input_columns.energy_demand_per_use}{_shape.input_rows.cooking.data}",
                equip.energy_demand_per_use,
            ),
            (
                f"{_shape.input_columns.selection}{_shape.input_rows.cooking.selection}",
                _shape.input_rows.cooking.selection_options[str(equip.cooktop_type)],
            ),
        ]
        return [xl_data.XlItem(_shape.name, *item) for item in items]

    def _mel(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _lighting_interior(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _lighting_exterior(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _lighting_garage(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _custom_elec(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _custom_lighting(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []

    def _custom_mel(self, _shape: shape_model.Electricity) -> List[xl_data.XlItem]:
        return []


# -----------------------------------------------------------------------------
# -- Worksheet Reader


class PHPPReadAddress(NamedTuple):
    """Attribute Name / PHPP Address pair."""

    attr_name: str
    phpp_address: str


class ReaderDataItem(NamedTuple):
    """Electric-Equipment data for a single PHPP item."""

    type: ElectricEquipmentType
    data: List[PHPPReadAddress]


class ReaderAddressesGroup(NamedTuple):
    """Electric-Equipment data for a single PHPP item."""

    used: str
    inside: str
    demand: str
    selection: str


@dataclass
class ElectricityItemXLReader:
    """Model class for defining read-locations for Electric-Equipment data in the PHPP."""

    __slots__ = ("shape", "cols", "rows")
    shape: shape_model.Electricity

    def __post_init__(self) -> None:
        """Aliases, just for convenience."""
        self.cols = self.shape.input_columns
        self.rows = self.shape.input_rows

    def _get_addresses(self, _rows: shape_model.ElectricityInputRow) -> ReaderAddressesGroup:
        """Return a group of addresses for a single PHPP Device."""
        return ReaderAddressesGroup(
            f"{self.cols.used}{_rows.data}",
            f"{self.cols.in_conditioned_space}{_rows.data}",
            f"{self.cols.annual_energy_demand}{_rows.data}",
            f"{self.cols.selection}{_rows.selection}",
        )

    @property
    def _dishwasher(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.dishwasher)
        return ReaderDataItem(
            ElectricEquipmentType.DISHWASHER,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
                PHPPReadAddress("water_connection", address.selection),
            ],
        )

    @property
    def _clothes_washer(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.clothes_washing)
        return ReaderDataItem(
            ElectricEquipmentType.CLOTHES_WASHER,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
                PHPPReadAddress("water_connection", address.selection),
            ],
        )

    @property
    def _clothes_dryer(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.clothes_drying)
        return ReaderDataItem(
            ElectricEquipmentType.CLOTHES_DRYER,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
                PHPPReadAddress("dryer_type", address.selection),
            ],
        )

    @property
    def _refrigerator(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.refrigerator)
        return ReaderDataItem(
            ElectricEquipmentType.REFRIGERATOR,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )

    @property
    def _fridge_freezer(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.fridge_freezer)
        return ReaderDataItem(
            ElectricEquipmentType.FRIDGE_FREEZER,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )

    @property
    def _freezer(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.freezer)
        return ReaderDataItem(
            ElectricEquipmentType.FREEZER,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )

    @property
    def _cooktop(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.cooking)
        return ReaderDataItem(
            ElectricEquipmentType.COOKING,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )

    @property
    def _mel(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.small_appliances)
        return ReaderDataItem(
            ElectricEquipmentType.MEL,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )

    @property
    def _lighting_interior(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.lighting_interior)
        return ReaderDataItem(
            ElectricEquipmentType.LIGHTING_INTERIOR,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )

    @property
    def _lighting_exterior(self) -> ReaderDataItem:
        address = self._get_addresses(self.rows.lighting_exterior)
        return ReaderDataItem(
            ElectricEquipmentType.LIGHTING_EXTERIOR,
            [
                PHPPReadAddress("quantity", address.used),
                PHPPReadAddress("in_conditioned_space", address.inside),
                PHPPReadAddress("energy_demand", address.demand),
            ],
        )



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Controller Class for the PHPP Shading worksheet."""

from __future__ import annotations

from typing import List, Optional

from PHX.PHPP.phpp_localization import shape_model
from PHX.PHPP.phpp_model import shading_rows
from PHX.xl import xl_app


class Shading:
    """IO Controller Class for PHPP "Shading" worksheet."""

    _header_row: Optional[int] = None
    _entry_row_start: Optional[int] = None
    _entry_row_end: Optional[int] = None

    def __init__(self, _xl: xl_app.XLConnection, shape: shape_model.Shading):
        self.xl = _xl
        self.shape = shape

    @property
    def header_row(self) -> int:
        """The row number for the Window entry 'Header'."""
        if not self._header_row:
            self._header_row = self.find_header_row()
        return self._header_row

    @property
    def entry_row_start(self) -> int:
        if not self._entry_row_start:
            self._entry_row_start = self.find_entry_block_start()
        return self._entry_row_start

    @property
    def entry_row_end(self) -> int:
        if not self._entry_row_end:
            self._entry_row_end = self.find_entry_block_end()
        return self._entry_row_end

    def find_header_row(self, _row_start: int = 1, _row_end: int = 100) -> int:
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.shading_rows.locator_col_header,
            _row_start=_row_start,
            _row_end=_row_end,
        )

        for i, val in enumerate(xl_data):
            if self.shape.shading_rows.locator_string_header == val:
                return i

        raise Exception(
            f"Error: Cannot find the '{self.shape.shading_rows.locator_string_header}' "
            f"header on the '{self.shape.name}' sheet, column {self.shape.shading_rows.locator_string_header}?"
        )

    def find_entry_block_start(self, _start_row: int = 1, _read_length: int = 100) -> int:
        """Return the starting row for the window data entry block."""

        end_row = _start_row + _read_length
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.shading_rows.locator_col_entry,
            _row_start=_start_row,
            _row_end=end_row,
        )

        for i, val in enumerate(xl_data, start=_start_row):
            if self.shape.shading_rows.locator_string_entry in str(val):
                return i + 2

        if end_row < 10_000:
            return self.find_entry_block_start(_start_row=end_row, _read_length=1_000)

        raise Exception(
            f"Error: Cannot find the '{self.shape.shading_rows.locator_string_entry}' "
            f"marker on the '{self.shape.name}' sheet, column {self.shape.shading_rows.locator_col_entry}?"
        )

    def find_entry_block_end(self, _start_row: int = 1, _read_length: int = 100) -> int:
        """Return the ending row for the window data entry block."""

        end_row = _start_row + _read_length
        xl_data = self.xl.get_single_column_data(
            _sheet_name=self.shape.name,
            _col=self.shape.shading_rows_end.locator_col_entry,
            _row_start=_start_row,
            _row_end=end_row,
        )

        for i, val in enumerate(xl_data, start=_start_row):
            if self.shape.shading_rows_end.locator_string_entry in str(val):
                return i + 2

        if end_row < 10_000:
            return self.find_entry_block_end(_start_row=end_row, _read_length=1_000)

        raise Exception(
            f"Error: Cannot find the '{self.shape.shading_rows_end.locator_string_entry}' "
            f"marker on the '{self.shape.name}' sheet, column {self.shape.shading_rows_end.locator_col_entry}?"
        )

    def write_shading(self, _shading_rows: List[shading_rows.ShadingRow]) -> None:
        """Write a list of ShadingRow objects to the Shading worksheet."""

        for i, shading_row in enumerate(_shading_rows, start=self.entry_row_start):
            for item in shading_row.create_xl_items(self.shape.name, _row_num=i):
                self.xl.write_xl_item(item)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Model class for a PHPP Areas / Thermal Bridges-Entry row"""

from dataclasses import dataclass
from functools import partial
from typing import List, Tuple

from PHX.model import components
from PHX.PHPP.phpp_localization import shape_model
from PHX.xl import xl_data


@dataclass
class ThermalBridgeRow:
    """A single Areas/Thermal Bridge entry row."""

    __slots__ = ("shape", "phx_tb")
    shape: shape_model.Areas
    phx_tb: components.PhxComponentThermalBridge

    def _create_range(self, _field_name: str, _row_num: int) -> str:
        """Return the XL Range ("P12",...) for the specific field name."""
        return f"{getattr(self.shape.thermal_bridge_rows.inputs, _field_name).column}{_row_num}"

    def _get_target_unit(self, _field_name: str) -> str:
        "Return the right target unit for the PHPP item writing (IP | SI)"
        return getattr(self.shape.thermal_bridge_rows.inputs, _field_name).unit

    def create_xl_items(self, _sheet_name: str, _row_num: int) -> List[xl_data.XlItem]:
        """Returns a list of the XL Items to write for this Thermal Bridge Entry

        Arguments:
        ----------
            * _sheet_name: (str) The name of the worksheet to write to.
            * _row_num: (int) The row number to build the XlItems for
        Returns:
        --------
            * (List[XlItem]): The XlItems to write to the sheet.
        """
        create_range = partial(self._create_range, _row_num=_row_num)
        XLItemAreas = partial(xl_data.XlItem, _sheet_name)
        return [
            XLItemAreas(create_range("description"), f"'{self.phx_tb.display_name}"),
            XLItemAreas(create_range("group_number"), self.phx_tb.group_type.value),
            XLItemAreas(create_range("quantity"), self.phx_tb.quantity),
            XLItemAreas(
                create_range("length"),
                self.phx_tb.length,
                "M",
                self._get_target_unit("length"),
            ),
            XLItemAreas(
                create_range("psi_value"),
                self.phx_tb.psi_value,
                "W/MK",
                self._get_target_unit("psi_value"),
            ),
            XLItemAreas(create_range("fRsi_value"), self.phx_tb.fRsi_value),
        ]



from typing import Any, Dict

from PHX.from_WUFI_XML.phx_schemas import _PhxProject
from PHX.from_WUFI_XML.wufi_file_schema import WUFIplusProject
from PHX.model import project


def convert_WUFI_XML_to_PHX_project(
    _wufi_xml_project: WUFIplusProject,
) -> project.PhxProject:
    """Convert a WUFI-XML-Project object over to a PHX-Project."""
    phx_project = _PhxProject(_wufi_xml_project)
    return phx_project



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions to create PHX-Service Hot Water objects from Honeybee-PH-HVAC Hot-Water"""

from honeybee_phhvac import hot_water_devices, hot_water_piping

from PHX.model import hvac
from PHX.model.enums.hvac import PhxHotWaterPipingInchDiameterType, PhxHotWaterPipingMaterial, PhxHotWaterTankType
from PHX.model.hvac import piping

# -- Storage ------------------------------------------------------------------


def build_phx_hw_tank(_ph_hvac_tank: hot_water_devices.PhHvacHotWaterTank) -> hvac.PhxHotWaterTank:
    """Returns a new PHX Hot-Water Tank based on the Honeybee-PH-HVAC Hot Water Tank input.

    Arguments:
    ----------
        * _ph_hvac_tank: The Honeybee-PH-HVAC Hot-Water tank to use as the source.

    Returns:
    --------
        * The new PHX-Hot-Water-Tank.
    """

    phx_tank = hvac.PhxHotWaterTank()

    phx_tank.display_name = _ph_hvac_tank.display_name
    phx_tank.quantity = _ph_hvac_tank.quantity

    phx_tank.params.tank_type = PhxHotWaterTankType.from_hbph_type(_ph_hvac_tank.tank_type)
    phx_tank.params.in_conditioned_space = _ph_hvac_tank.in_conditioned_space

    phx_tank.params.solar_connection = _ph_hvac_tank.solar_connection
    phx_tank.params.solar_losses = _ph_hvac_tank.solar_losses

    phx_tank.params.storage_capacity = _ph_hvac_tank.storage_capacity
    phx_tank.params.storage_loss_rate = _ph_hvac_tank.storage_loss_rate

    phx_tank.params.standby_losses = _ph_hvac_tank.standby_losses
    phx_tank.params.standby_fraction = _ph_hvac_tank.standby_fraction

    phx_tank.params.room_temp = _ph_hvac_tank.room_temp
    phx_tank.params.water_temp = _ph_hvac_tank.water_temp

    return phx_tank


def build_phx_hw_storage(_ph_hvac_tank: hot_water_devices.PhHvacHotWaterTank) -> hvac.PhxHotWaterTank:
    """Returns a new PHX Hot-Water Storage Tank based on the Honeybee-PH-HVAC Hot Water Tank input.

    Arguments:
    ----------
        * _ph_hvac_tank: The Honeybee-PH-HVAC Hot-Water tank to use as the source.

    Returns:
    --------
        * The new Water Storage Tank.
    """

    phx_storage_tank = build_phx_hw_tank(_ph_hvac_tank)

    return phx_storage_tank


# -- Heaters ------------------------------------------------------------------


def build_phx_hw_heater(
    _ph_hvac_heater: hot_water_devices.PhHvacHotWaterHeater,
) -> hvac.AnyPhxHotWaterHeater:
    """Returns a new PHX Hot-Water Heater based on the Honeybee-PH-HVAC Hot Water Heater input.

    Arguments:
    ----------
        * _ph_hvac_heater: The Honeybee-PH-HVAC Hot-Water heater to use as the source for the PHX Heater.

    Returns:
    --------
        * The new PHX-Hot-Water-Heater.
    """

    # -- Get the right constructor based on the type of Honeybee-PH-HVAC heater
    heaters = {
        "PhHvacHotWaterHeaterElectric": hvac.PhxHeaterElectric,
        "PhHvacHotWaterHeaterBoiler": hvac.PhxHeaterBoilerFossil,
        "PhHvacHotWaterHeaterBoilerWood": hvac.PhxHeaterBoilerWood,
        "PhHvacHotWaterHeaterDistrict": hvac.PhxHeaterDistrictHeat,
        "PhHvacHotWaterHeaterHeatPump_Annual": hvac.PhxHeatPumpAnnual,
        "PhHvacHotWaterHeaterHeatPump_Monthly": hvac.PhxHeatPumpMonthly,
        "PhHvacHotWaterHeaterHeatPump_Inside": hvac.PhxHeatPumpHotWater,
    }

    # -- Build the basic heater and set basic data
    heater_class = heaters[_ph_hvac_heater.__class__.__name__]
    phx_hw_heater = heater_class()
    phx_hw_heater.display_name = _ph_hvac_heater.display_name

    # -- Pull out all the detailed data which varies depending on the 'type'
    for attr_name in vars(phx_hw_heater).keys():
        try:
            if attr_name.startswith("_"):
                attr_name = attr_name[1:]
            setattr(phx_hw_heater, attr_name, getattr(_ph_hvac_heater, attr_name))
        except AttributeError:
            pass

    for attr_name in vars(phx_hw_heater.params).keys():
        try:
            if attr_name.startswith("_"):
                attr_name = attr_name[1:]
            setattr(phx_hw_heater.params, attr_name, getattr(_ph_hvac_heater, attr_name))
        except AttributeError:
            pass

    phx_hw_heater.usage_profile.dhw_heating = True

    return phx_hw_heater


# -- Piping -------------------------------------------------------------------


def build_phx_pipe_element(_ph_hvac_pipe: hot_water_piping.PhHvacPipeElement) -> hvac.PhxPipeElement:
    """Create single PHX-PipeElement based on an input Honeybee-PH-HVAC PipeElement.

    Arguments:
    ----------
        * _ph_hvac_pipe: The Honeybee-PH-HVAC Pipe to use as the source.

    Returns:
    --------
        * The new PHX Pipe Element created.
    """

    phx_pipe = piping.PhxPipeElement(_ph_hvac_pipe.identifier, _ph_hvac_pipe.display_name)
    for segment in _ph_hvac_pipe.segments:
        phx_pipe.add_segment(
            piping.PhxPipeSegment(
                segment.identifier,
                segment.display_name,
                segment.geometry,
                PhxHotWaterPipingMaterial(segment.material.number),
                segment.diameter_m,
                segment.insulation_thickness_m,
                segment.insulation_conductivity,
                segment.insulation_reflective,
                segment.insulation_quality,
                segment.daily_period,
                segment.water_temp_c,
            )
        )

    return phx_pipe


def build_phx_branch_pipe(_ph_hvac_branch: hot_water_piping.PhHvacPipeBranch) -> hvac.PhxPipeBranch:
    """Create new PHX-Branch-Pipe based on an input Honeybee-PH-HVAC Branch-Pipe.

    Arguments:
    ----------
        * _ph_hvac_branch: The Honeybee-PH-HVAC Branch-Pipe to use as the source.

    Returns:
    --------
        * The new PHX-Branch created.
    """

    phx_pipe_branch = piping.PhxPipeBranch(_ph_hvac_branch.identifier, _ph_hvac_branch.display_name)
    for segment in _ph_hvac_branch.segments:
        phx_pipe_branch.pipe_element.add_segment(
            piping.PhxPipeSegment(
                segment.identifier,
                segment.display_name,
                segment.geometry,
                PhxHotWaterPipingMaterial(segment.material.number),
                segment.diameter_m,
                segment.insulation_thickness_m,
                segment.insulation_conductivity,
                segment.insulation_reflective,
                segment.insulation_quality,
                segment.daily_period,
                segment.water_temp_c,
            )
        )

    for fixture in _ph_hvac_branch.fixtures:
        phx_pipe_branch.add_fixture(build_phx_pipe_element(fixture))

    return phx_pipe_branch


def build_phx_trunk_pipe(_ph_hvac_trunk: hot_water_piping.PhHvacPipeTrunk) -> hvac.PhxPipeTrunk:
    """Create new PHX-Trunk based on an input Honeybee-PH-HVAC Trunk-Pipe.

    Arguments:
    ----------
        * _ph_hvac_trunk: The Honeybee-PH-HVAC Trunk-Pipe to use as the source.

    Returns:
    --------
        * The new PHX-Trunk created.
    """

    phx_pipe_trunk = piping.PhxPipeTrunk(_ph_hvac_trunk.identifier, _ph_hvac_trunk.display_name)
    phx_pipe_trunk.multiplier = _ph_hvac_trunk.multiplier
    phx_pipe_trunk.pipe_element.display_name = f"{_ph_hvac_trunk.display_name}-PIPE"
    for segment in _ph_hvac_trunk.segments:
        phx_pipe_trunk.pipe_element.add_segment(
            piping.PhxPipeSegment(
                segment.identifier,
                segment.display_name,
                segment.geometry,
                PhxHotWaterPipingMaterial(segment.material.number),
                segment.diameter_m,
                segment.insulation_thickness_m,
                segment.insulation_conductivity,
                segment.insulation_reflective,
                segment.insulation_quality,
                segment.daily_period,
                segment.water_temp_c,
            )
        )

    for branch in _ph_hvac_trunk.branches:
        phx_pipe_trunk.add_branch(build_phx_branch_pipe(branch))

    return phx_pipe_trunk



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Mechanical Ventilation Devices"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, Optional, Union

from PHX.model.enums.hvac import DeviceType, PhxExhaustVentType, SystemType
from PHX.model.hvac import _base


@dataclass
class PhxDeviceVentilation(_base.PhxMechanicalDevice):
    def __post_init__(self):
        super().__post_init__()
        self.usage_profile.ventilation = True


# -- HRV / ERV Air  -----------------------------------------------------------


@dataclass
class PhxDeviceVentilatorParams(_base.PhxMechanicalDeviceParams):
    _sensible_heat_recovery: float = 0.0
    _latent_heat_recovery: float = 0.0
    _quantity: int = 1
    _electric_efficiency: float = 0.55
    _frost_protection_reqd: bool = True
    _temperature_below_defrost_used: float = -5.0

    @property
    def sensible_heat_recovery(self) -> float:
        return self._sensible_heat_recovery

    @sensible_heat_recovery.setter
    def sensible_heat_recovery(self, value: Optional[float]) -> None:
        if value is not None:
            self._sensible_heat_recovery = value

    @property
    def latent_heat_recovery(self) -> float:
        return self._latent_heat_recovery

    @latent_heat_recovery.setter
    def latent_heat_recovery(self, value: Optional[float]) -> None:
        if value is not None:
            self._latent_heat_recovery = value

    @property
    def quantity(self) -> int:
        return self._quantity

    @quantity.setter
    def quantity(self, value: Optional[int]) -> None:
        if value is not None:
            self._quantity = value

    @property
    def electric_efficiency(self) -> float:
        return self._electric_efficiency

    @electric_efficiency.setter
    def electric_efficiency(self, value: Optional[float]) -> None:
        if value is not None:
            self._electric_efficiency = value

    @property
    def frost_protection_reqd(self) -> bool:
        return self._frost_protection_reqd

    @frost_protection_reqd.setter
    def frost_protection_reqd(self, value: Optional[bool]) -> None:
        if value is not None:
            self._frost_protection_reqd = value

    @property
    def temperature_below_defrost_used(self) -> float:
        return self._temperature_below_defrost_used

    @temperature_below_defrost_used.setter
    def temperature_below_defrost_used(self, value: Optional[float]) -> None:
        if value is not None:
            self._temperature_below_defrost_used = value

    def __add__(self, other: PhxDeviceVentilatorParams) -> PhxDeviceVentilatorParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.sensible_heat_recovery = (self.sensible_heat_recovery + other.sensible_heat_recovery) / 2
        new_obj.latent_heat_recovery = (self.latent_heat_recovery + other.latent_heat_recovery) / 2
        new_obj.quantity = self.quantity + other.quantity
        new_obj.electric_efficiency = (self.electric_efficiency + other.electric_efficiency) / 2
        new_obj.frost_protection_reqd = any([self.frost_protection_reqd, other.frost_protection_reqd])
        new_obj.temperature_below_defrost_used = (
            self.temperature_below_defrost_used + other.temperature_below_defrost_used
        ) / 2
        return new_obj


@dataclass
class PhxDeviceVentilator(PhxDeviceVentilation):
    system_type: SystemType = field(init=False, default=SystemType.VENTILATION)
    device_type: DeviceType = field(init=False, default=DeviceType.VENTILATION)
    params: PhxDeviceVentilatorParams = field(default_factory=PhxDeviceVentilatorParams)

    def __post_init__(self):
        super().__post_init__()

    def __add__(self, other: PhxDeviceVentilator) -> PhxDeviceVentilator:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        return new_obj


# -- Exhaust Ventilators (Kitchen/Dryer) --------------------------------------


@dataclass
class PhxExhaustVentilatorParams(_base.PhxMechanicalDeviceParams):
    _exhaust_type: PhxExhaustVentType = PhxExhaustVentType.KITCHEN_HOOD
    _annual_runtime_minutes: float = 0.0
    _exhaust_flow_rate_m3h: float = 0.0

    @property
    def exhaust_type(self) -> PhxExhaustVentType:
        return self._exhaust_type

    @exhaust_type.setter
    def exhaust_type(self, value: Optional[PhxExhaustVentType]) -> None:
        if value is not None:
            self._exhaust_type = value

    @property
    def annual_runtime_minutes(self) -> float:
        return self._annual_runtime_minutes

    @annual_runtime_minutes.setter
    def annual_runtime_minutes(self, value: Optional[float]) -> None:
        if value is not None:
            self._annual_runtime_minutes = value

    @property
    def exhaust_flow_rate_m3h(self) -> float:
        return self._exhaust_flow_rate_m3h

    @exhaust_flow_rate_m3h.setter
    def exhaust_flow_rate_m3h(self, value: Optional[float]) -> None:
        if value is not None:
            self._exhaust_flow_rate_m3h = value

    def _calc_flow_weighted_annual_minutes(self, other: PhxExhaustVentilatorParams) -> float:
        """Return the flow-weighted annual runtime minutes.

        ie: Device A (50 mins @ 450 m3/h) + Device B (40 mins @ 200 m3/h) -> 47 mins
        """
        weighted_flow_a = self.annual_runtime_minutes * self.exhaust_flow_rate_m3h
        weighted_flow_b = other.annual_runtime_minutes * other.exhaust_flow_rate_m3h
        total_weighted_flow = weighted_flow_a + weighted_flow_b
        total_flow = self.exhaust_flow_rate_m3h + other.exhaust_flow_rate_m3h
        weighted_minutes = total_weighted_flow / total_flow
        return weighted_minutes

    def __add__(self, other: PhxExhaustVentilatorParams) -> PhxExhaustVentilatorParams:
        if not self.exhaust_type == other.exhaust_type:
            msg = (
                f"Error: Cannot combine PHX Exhaust Ventilation Device of type "
                f"'{self.exhaust_type}' with device of type: '{other.exhaust_type}'"
            )
            raise Exception(msg)

        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))

        if (self.annual_runtime_minutes > 0) or (other.annual_runtime_minutes > 0):
            new_obj.annual_runtime_minutes = self._calc_flow_weighted_annual_minutes(other)

        new_obj.exhaust_flow_rate_m3h = self.exhaust_flow_rate_m3h + other.exhaust_flow_rate_m3h
        return new_obj


@dataclass
class PhxExhaustVentilatorBase(_base.PhxMechanicalDevice):
    """Base class for all Exhaust Ventilation.

    Note that the _count and id_num are implemented here to allow each unique
    type of device to increment the count at the base-class level without affecting
    the _base.PhxMechanicalDevice class.
    """

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)

    def __post_init__(self):
        super().__post_init__()
        # -- Use PhxExhaustVentilatorBase._count instead of super() so that all
        # -- counting happens at the parent class, not the child class.
        PhxExhaustVentilatorBase._count += 1
        self.id_num = PhxExhaustVentilatorBase._count
        self.quantity = 1


@dataclass
class PhxExhaustVentilatorRangeHood(PhxExhaustVentilatorBase):
    system_type: SystemType = field(init=False, default=SystemType.VENTILATION)
    device_type: DeviceType = field(init=False, default=DeviceType.VENTILATION)
    params: PhxExhaustVentilatorParams = field(default_factory=PhxExhaustVentilatorParams)

    def __post_init__(self):
        super().__post_init__()
        self.params.exhaust_type = PhxExhaustVentType.KITCHEN_HOOD

    def __add__(self, other: PhxExhaustVentilatorRangeHood) -> PhxExhaustVentilatorRangeHood:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.quantity = self.quantity + other.quantity
        new_obj.display_name = f"Kitchen Hoods ({new_obj.quantity})"

        return new_obj

    def __str__(self):
        return f"{self.__class__.__name__}({self.display_name}, {self.params.exhaust_flow_rate_m3h:,.1f}m3/h)"


@dataclass
class PhxExhaustVentilatorDryer(PhxExhaustVentilatorBase):
    system_type: SystemType = field(init=False, default=SystemType.VENTILATION)
    device_type: DeviceType = field(init=False, default=DeviceType.VENTILATION)
    params: PhxExhaustVentilatorParams = field(default_factory=PhxExhaustVentilatorParams)

    def __post_init__(self):
        PhxExhaustVentilatorBase.__post_init__(self)
        self.params.exhaust_type = PhxExhaustVentType.DRYER

    def __add__(self, other: PhxExhaustVentilatorDryer) -> PhxExhaustVentilatorDryer:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.quantity = self.quantity + other.quantity
        new_obj.display_name = f"Dryers ({new_obj.quantity})"
        return new_obj

    def __str__(self):
        return f"{self.__class__.__name__}({self.display_name}, {self.params.exhaust_flow_rate_m3h:,.1f}m3/h)"


@dataclass
class PhxExhaustVentilatorUserDefined(PhxExhaustVentilatorBase):
    system_type: SystemType = field(init=False, default=SystemType.VENTILATION)
    device_type: DeviceType = field(init=False, default=DeviceType.VENTILATION)
    params: PhxExhaustVentilatorParams = field(default_factory=PhxExhaustVentilatorParams)

    def __post_init__(self):
        super().__post_init__()
        self.params.exhaust_type = PhxExhaustVentType.USER_DEFINED

    def __add__(self, other: PhxExhaustVentilatorUserDefined) -> PhxExhaustVentilatorUserDefined:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.quantity = self.quantity + other.quantity
        new_obj.display_name = f"User-Determined ({new_obj.quantity})"
        return new_obj

    def __str__(self):
        return f"{self.__class__.__name__}({self.display_name}, {self.params.exhaust_flow_rate_m3h:,.1f}m3/h)"


# -----------------------------------------------------------------------------
# ---- Type Alias


AnyPhxVentilation = PhxDeviceVentilator

AnyPhxExhaustVent = Union[
    PhxExhaustVentilatorRangeHood,
    PhxExhaustVentilatorDryer,
    PhxExhaustVentilatorUserDefined,
]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Passive House Electrical Equipment (Appliances) Classes"""

import inspect
import sys
import uuid
from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional, Type, Union

from PHX.model.enums.elec_equip import ElectricEquipmentType


@dataclass
class PhxElectricalDevice:
    """Base class for PHX Electrical Equipment (dishwashers, laundry, lighting, etc.)"""

    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)
    identifier: Union[uuid.UUID, str] = field(default_factory=uuid.uuid4)

    _reference_quantity: int = 1
    _reference_energy_norm: int = 2
    _quantity: int = 1

    display_name: Optional[str] = "_unnamed_equipment_"
    comment: Optional[str] = ""
    in_conditioned_space: Optional[bool] = True
    energy_demand: Optional[float] = 100.0
    energy_demand_per_use: Optional[float] = 100.0
    combined_energy_factor: Optional[float] = 0.0

    device_type: ElectricEquipmentType = field(default=ElectricEquipmentType.CUSTOM)

    def __post_init__(self) -> None:
        self.__class__._count += 1
        self.id_num = self.__class__._count

    @property
    def reference_quantity(self) -> int:
        return self._reference_quantity

    @reference_quantity.setter
    def reference_quantity(self, value: Optional[int]) -> None:
        if value is not None:
            self._reference_quantity = value

    @property
    def reference_energy_norm(self) -> int:
        return self._reference_energy_norm

    @reference_energy_norm.setter
    def reference_energy_norm(self, value: Optional[int]) -> None:
        if value is not None:
            self._reference_energy_norm = value

    @property
    def quantity(self) -> int:
        return self._quantity

    @quantity.setter
    def quantity(self, value: Optional[int]) -> None:
        if value is not None:
            self._quantity = value

    def get_energy_demand(self) -> float:
        """To allow for subclass custom behavior. Cannot use @property since
        it will not work with __setattr__ which is used during HBPH->PHX object creation.
        """
        return self.energy_demand or 0.0

    def get_quantity(self) -> int:
        """To allow for subclass custom behavior. Cannot use @property since
        it will not work with __setattr__ which is used during HBPH->PHX object creation.
        """
        return self.quantity or 1


class PhxDeviceDishwasher(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self._water_connection: int = 1  # DHW Connection
        self._capacity_type: int = 1
        self.display_name = "Kitchen Dishwasher"
        self.capacity: Optional[float] = 1.0
        self.device_type = ElectricEquipmentType.DISHWASHER

    @property
    def water_connection(self) -> Optional[int]:
        return self._water_connection

    @water_connection.setter
    def water_connection(self, value: Optional[int]) -> None:
        if value is not None:
            self._water_connection = value

    @property
    def capacity_type(self) -> Optional[int]:
        return self._capacity_type

    @capacity_type.setter
    def capacity_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._capacity_type = value


class PhxDeviceClothesWasher(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "Laundry - washer"
        self._water_connection: int = 1  # DHW Connection

        self.capacity: Optional[float] = 0.0814  # m3
        self.modified_energy_factor: Optional[float] = 2.38
        self.utilization_factor: Optional[float] = 1.0
        self.device_type = ElectricEquipmentType.CLOTHES_WASHER

    @property
    def water_connection(self) -> Optional[int]:
        return self._water_connection

    @water_connection.setter
    def water_connection(self, value: Optional[int]) -> None:
        if value is not None:
            self._water_connection = value


class PhxDeviceClothesDryer(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "Laundry - dryer"
        self._dryer_type: int = 4  # Condensation dryer
        self._field_utilization_factor_type: int = 1  # Timer

        self.gas_consumption: Optional[float] = 0.0  # kWh
        self.gas_efficiency_factor: Optional[float] = 2.67
        self.field_utilization_factor: Optional[float] = 1.18
        self.device_type = ElectricEquipmentType.CLOTHES_DRYER

    @property
    def dryer_type(self) -> int:
        return self._dryer_type

    @dryer_type.setter
    def dryer_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._dryer_type = value

    @property
    def field_utilization_factor_type(self) -> Optional[int]:
        return self._field_utilization_factor_type

    @field_utilization_factor_type.setter
    def field_utilization_factor_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._field_utilization_factor_type = value


class PhxDeviceRefrigerator(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "Kitchen refrigerator"
        self.device_type = ElectricEquipmentType.REFRIGERATOR


class PhxDeviceFreezer(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "kitchen freezer"
        self.device_type = ElectricEquipmentType.FREEZER


class PhxDeviceFridgeFreezer(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "Kitchen fridge/freeze combo"
        self.device_type = ElectricEquipmentType.FRIDGE_FREEZER


class PhxDeviceCooktop(PhxElectricalDevice):
    def __init__(self):
        super().__init__()
        self.display_name = "Kitchen cooking"
        self._cooktop_type: int = 1  # Electric
        self.device_type = ElectricEquipmentType.COOKING

    @property
    def cooktop_type(self) -> int:
        return self._cooktop_type

    @cooktop_type.setter
    def cooktop_type(self, value: Optional[int]) -> None:
        if value is not None:
            self._cooktop_type = value


class PhxDeviceMEL(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "PHIUS+ MELS"
        self.device_type = ElectricEquipmentType.MEL


class PhxDeviceLightingInterior(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "PHIUS+ Interior Lighting"
        self.frac_high_efficiency: Optional[float] = 1.0
        self.device_type = ElectricEquipmentType.LIGHTING_INTERIOR


class PhxDeviceLightingExterior(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "PHIUS+ Exterior Lighting"
        self.frac_high_efficiency: Optional[float] = 1.0
        self.device_type = ElectricEquipmentType.LIGHTING_EXTERIOR
        self.in_conditioned_space = False


class PhxDeviceLightingGarage(PhxElectricalDevice):
    def __init__(self) -> None:
        super().__init__()
        self.display_name = "PHIUS+ Garage Lighting"
        self.frac_high_efficiency: Optional[float] = 1.0
        self.device_type = ElectricEquipmentType.LIGHTING_GARAGE
        self.in_conditioned_space = False


class PhxDeviceCustomElec(PhxElectricalDevice):
    def __init__(self) -> None:
        self.display_name = "User defined"
        super().__init__()
        self.device_type = ElectricEquipmentType.CUSTOM


class PhxDeviceCustomLighting(PhxElectricalDevice):
    """Override so that WUFI output quantity shows up as 1"""

    def __init__(self) -> None:
        self.display_name = "User defined - lighting"
        super().__init__()
        self.device_type = ElectricEquipmentType.CUSTOM_LIGHTING

    def get_energy_demand(self) -> float:
        if self.energy_demand is None:
            return 0
        return self.energy_demand * self.quantity

    def get_quantity(self) -> int:
        return 1


class PhxDeviceCustomMEL(PhxElectricalDevice):
    """Override so that WUFI output quantity shows up as 1"""

    def __init__(self) -> None:
        self.display_name = "User defined - Misc electrical loads"
        super().__init__()
        self.device_type = ElectricEquipmentType.CUSTOM_MEL

    def get_energy_demand(self) -> float:
        if self.energy_demand is None:
            return 0
        return self.energy_demand * self.quantity

    def get_quantity(self) -> int:
        return 1


class PhxElevatorHydraulic(PhxElectricalDevice):
    def __init__(self) -> None:
        self.display_name = "User defined - Misc electrical loads"
        super().__init__()
        self.device_type = ElectricEquipmentType.CUSTOM

    def get_energy_demand(self) -> float:
        if self.energy_demand is None:
            return 0
        return self.energy_demand * self.quantity

    def get_quantity(self) -> int:
        return 1


class PhxElevatorGearedTraction(PhxElectricalDevice):
    def __init__(self) -> None:
        self.display_name = "User defined - Misc electrical loads"
        super().__init__()
        self.device_type = ElectricEquipmentType.CUSTOM

    def get_energy_demand(self) -> float:
        if self.energy_demand is None:
            return 0
        return self.energy_demand * self.quantity

    def get_quantity(self) -> int:
        return 1


class PhxElevatorGearlessTraction(PhxElectricalDevice):
    def __init__(self) -> None:
        self.display_name = "User defined - Misc electrical loads"
        super().__init__()
        self.device_type = ElectricEquipmentType.CUSTOM

    def get_energy_demand(self) -> float:
        if self.energy_demand is None:
            return 0
        return self.energy_demand * self.quantity

    def get_quantity(self) -> int:
        return 1


# -----------------------------------------------------------------------------


def get_device_type_map() -> Dict[ElectricEquipmentType, Type[PhxElectricalDevice]]:
    """Returns a dictionary mapping all the device enum types to the PhxElectricalDevice classes.

    This is useful of you are building up new Devices from raw data or types inputs.
    """

    d = {}
    for name, device_class in inspect.getmembers(sys.modules[__name__]):
        if not inspect.isclass(device_class):
            continue
        if not issubclass(device_class, PhxElectricalDevice):
            continue

        obj = device_class()
        d[obj.device_type] = device_class

    return d


# -----------------------------------------------------------------------------


@dataclass
class PhxElectricDeviceCollection:
    """A collection of all the PhxElectricalDevices (laundry, lighting, etc.) in the Zone"""

    _devices: dict = field(default_factory=dict)

    @property
    def devices(self) -> List[PhxElectricalDevice]:
        """Returns a list of all the devices in the PhxElectricDeviceCollection, sorted by display_name."""
        if not self._devices:
            return []
        return sorted(self._devices.values(), key=lambda e: e.display_name or "")

    def device_key_in_collection(self, _device_key) -> bool:
        """Returns True if the key supplied is in the existing device collection."""
        return _device_key in self._devices.keys()

    def get_equipment_by_key(self, _key: str) -> Optional[PhxElectricalDevice]:
        return self._devices.get(_key, None)

    def add_new_device(self, _key: str, _device: PhxElectricalDevice) -> None:
        """Adds a new PHX Electric-Equipment device to the PhxElectricDeviceCollection.

        Arguments:
        ----------
            * _key (str): The key to use when storing the new electric-equipment
            * _devices (PhxElectricalDevice): The new PhxElectricalDevice to
                add to the collection.

        Returns:
        --------
            * None
        """
        self._devices[_key] = _device

    def __iter__(self):
        """Get each device in the PhxElectricDeviceCollection, one at a time."""
        for _ in self.devices:
            yield _

    def __len__(self) -> int:
        """Number of devices in the PhxElectricDeviceCollection"""
        return len(self.devices)

    def __bool__(self) -> bool:
        return bool(self._devices)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import pathlib

# -- Project information -----------------------------------------------------
# Add the root folder (up two from conf.py) to avoid Import Errors.
import sys

sys.path.insert(0, str(pathlib.Path(__file__).parents[3]))

# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "PHX"
copyright = "2022, bldgtyp, llc"
author = "PH-Tools"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.autodoc", "sphinx.ext.coverage", "sphinx.ext.napoleon"]


templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "alabaster"
html_static_path = ["_static"]



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Program Class for organizing Fresh-Air Ventilation Data."""

from __future__ import annotations

from dataclasses import dataclass, field

from PHX.model.loads import ventilation as vent_loads
from PHX.model.schedules import ventilation as vent_schedules


@dataclass
class PhxProgramVentilation:
    """A PHX Program for the Fresh-Air Ventilation with a load and schedule."""

    display_name: str = "Unnamed_Ventilation_Program"
    load: vent_loads.PhxLoadVentilation = field(default_factory=vent_loads.PhxLoadVentilation)
    schedule: vent_schedules.PhxScheduleVentilation = field(default_factory=vent_schedules.PhxScheduleVentilation)

    @property
    def has_ventilation_airflow(self) -> bool:
        """Returns True if the load has any amount of ventilation airflow."""
        return self.load.total_airflow > 0.0



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Aux-Energy . Supportive Extra Devices."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar

from PHX.model.enums.hvac import PhxSupportiveDeviceType
from PHX.model.hvac import _base


@dataclass
class PhxSupportiveDeviceParams(_base.PhxMechanicalDeviceParams):
    in_conditioned_space: bool = True
    norm_energy_demand_W: float = 1.0
    annual_period_operation_khrs: float = 8.760

    def __add__(self, other: PhxSupportiveDeviceParams) -> PhxSupportiveDeviceParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))
        new_obj.in_conditioned_space = any([self.in_conditioned_space, other.in_conditioned_space])

        # -- Merge the energy usage and hours of operation
        total_kilohours = self.annual_period_operation_khrs + other.annual_period_operation_khrs
        new_obj.annual_period_operation_khrs = total_kilohours
        total_watt_kilohours = (
            self.norm_energy_demand_W * self.annual_period_operation_khrs
            + other.norm_energy_demand_W * other.annual_period_operation_khrs
        )
        new_obj.norm_energy_demand_W = total_watt_kilohours / total_kilohours
        return new_obj


@dataclass
class PhxSupportiveDevice(_base.PhxMechanicalDevice):
    _count: ClassVar[int] = 0
    id_num: int = field(init=False, default=0)

    display_name: str = "__unnamed_device__"
    device_type: PhxSupportiveDeviceType = field(default=PhxSupportiveDeviceType.OTHER)
    quantity: int = 1
    params: PhxSupportiveDeviceParams = field(default_factory=PhxSupportiveDeviceParams)

    def __post_init__(self):
        super().__post_init__()
        PhxSupportiveDevice._count += 1
        self.id_num = PhxSupportiveDevice._count

    def __add__(self, other: PhxSupportiveDevice) -> PhxSupportiveDevice:
        if self.device_type != other.device_type:
            raise ValueError(f"Cannot add two different device types: {self.device_type} and {other.device_type}")
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        new_obj.display_name = f"Supportive Device ({new_obj.quantity})"

        return new_obj



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""PHX Water Devices"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from PHX.model.enums.hvac import DeviceType, PhxHotWaterInputOptions, PhxHotWaterTankType, SystemType
from PHX.model.hvac import _base


@dataclass
class PhxHotWaterDevice(_base.PhxMechanicalDevice):
    def __post_init__(self):
        super().__post_init__()
        self.usage_profile.dhw_heating = True


# -- Hot Water Tank -----------------------------------------------------------


@dataclass
class PhxHotWaterTankParams(_base.PhxMechanicalDeviceParams):
    # -- Device Params
    display_name: str = "_unnamed_PHX_hw_tank_"

    _tank_type: PhxHotWaterTankType = PhxHotWaterTankType.NONE
    _input_option: PhxHotWaterInputOptions = PhxHotWaterInputOptions.SPEC_TOTAL_LOSSES

    _in_conditioned_space: bool = True
    _solar_connection: bool = False
    _solar_losses: float = 0.0  # W/K
    _storage_loss_rate: float = 0.0  # W
    _storage_capacity: float = 0.0  # Liter
    _standby_losses: float = 4.0  # W/K
    _standby_fraction: float = 0.30  # %
    _room_temp: float = 20.0
    _water_temp: float = 60.0

    @property
    def tank_type(self) -> PhxHotWaterTankType:
        return self._tank_type

    @tank_type.setter
    def tank_type(self, value: Optional[PhxHotWaterTankType]) -> None:
        if value is not None:
            self._tank_type = value

    @property
    def input_option(self) -> PhxHotWaterInputOptions:
        return self._input_option

    @input_option.setter
    def input_option(self, value: Optional[PhxHotWaterInputOptions]) -> None:
        if value is not None:
            self._input_option = value

    @property
    def in_conditioned_space(self) -> bool:
        return self._in_conditioned_space

    @in_conditioned_space.setter
    def in_conditioned_space(self, value: Optional[bool]) -> None:
        if value is not None:
            self._in_conditioned_space = value

    @property
    def solar_connection(self) -> bool:
        return self._solar_connection

    @solar_connection.setter
    def solar_connection(self, value: Optional[bool]) -> None:
        if value is not None:
            self._solar_connection = value

    @property
    def solar_losses(self) -> float:
        return self._solar_losses

    @solar_losses.setter
    def solar_losses(self, value: Optional[float]) -> None:
        if value is not None:
            self._solar_losses = value

    @property
    def storage_loss_rate(self) -> float:
        return self._storage_loss_rate

    @storage_loss_rate.setter
    def storage_loss_rate(self, value: Optional[float]) -> None:
        if value is not None:
            self._storage_loss_rate = value

    @property
    def storage_capacity(self) -> float:
        return self._storage_capacity

    @storage_capacity.setter
    def storage_capacity(self, value: Optional[float]) -> None:
        if value is not None:
            self._storage_capacity = value

    @property
    def standby_losses(self) -> float:
        return self._standby_losses

    @standby_losses.setter
    def standby_losses(self, value: Optional[float]) -> None:
        if value is not None:
            self._standby_losses = value

    @property
    def standby_fraction(self) -> float:
        return self._standby_fraction

    @standby_fraction.setter
    def standby_fraction(self, value: Optional[float]) -> None:
        if value is not None:
            self._standby_fraction = value

    @property
    def room_temp(self) -> float:
        return self._room_temp

    @room_temp.setter
    def room_temp(self, value: Optional[float]) -> None:
        if value is not None:
            self._room_temp = value

    @property
    def water_temp(self) -> float:
        return self._water_temp

    @water_temp.setter
    def water_temp(self, value: Optional[float]) -> None:
        if value is not None:
            self._water_temp = value

    def __add__(self, other: PhxHotWaterTankParams) -> PhxHotWaterTankParams:
        base = super().__add__(other)
        new_obj = self.__class__(**vars(base))

        new_obj.tank_type = self.tank_type
        new_obj.input_option = self.input_option

        new_obj.in_conditioned_space = any([self.in_conditioned_space, other.in_conditioned_space])
        new_obj.solar_connection = any([self.solar_connection, other.solar_connection])
        new_obj.solar_losses = (self.solar_losses + other.solar_losses) / 2

        new_obj.storage_loss_rate = (self.storage_loss_rate + other.storage_loss_rate) / 2
        new_obj.storage_capacity = (self.storage_capacity + other.storage_capacity) / 2

        new_obj.standby_losses = (self.standby_losses + other.standby_losses) / 2
        new_obj.standby_fraction = (self.standby_fraction + other.standby_fraction) / 2

        new_obj.room_temp = (self.room_temp + other.room_temp) / 2
        new_obj.water_temp = (self.water_temp + other.water_temp) / 2

        return new_obj


@dataclass
class PhxHotWaterTank(PhxHotWaterDevice):
    system_type: SystemType = field(init=False, default=SystemType.WATER_STORAGE)
    device_type: DeviceType = field(init=False, default=DeviceType.WATER_STORAGE)
    params: PhxHotWaterTankParams = field(default_factory=PhxHotWaterTankParams)

    def __add__(self, other: PhxHotWaterTank) -> PhxHotWaterTank:
        base = super().__add__(other)
        new_obj = self.__class__.from_kwargs(**vars(base))
        return new_obj


AnyWaterTank = PhxHotWaterTank



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Basic datatypes and data-structures relevant for Excel read/write."""

import string
from typing import Dict, List, Optional, Tuple, Union

from ph_units import converter

xl_writable = Optional[Union[str, float, int, List, Tuple]]
xl_range_single_value = Union[str, float, int, None]
xl_range_list1D_value = Union[List[str], List[float], List[int], List[None]]
xl_range_list2D_value = Union[List[List[str]], List[List[float]], List[List[int]], List[List[None]]]
xl_range_value = Union[xl_range_single_value, xl_range_list1D_value, xl_range_list2D_value]


def xl_ord(_col: str) -> int:
    """ord() which supports excel columns beyond Z (AA, AB, ...)"""
    num = 0
    for c in _col.upper():
        if c in string.ascii_letters:
            num = num * 26 + (ord(c.upper()) - ord("A")) + 1
    return num + 64


def xl_chr(_i: int) -> str:
    """chr() which supports excel columns beyond Z (AA, AB, ...)"""
    letters = ""
    num = _i - 64
    while num:
        mod = (num - 1) % 26
        letters += chr(mod + 65)
        num = (num - 1) // 26
    return "".join(reversed(letters))


def xl_col_num_as_chr(_i: int) -> str:
    """Return the character for an Excel Column Num (1-based).
    ie: 1->"A", 2->"B", etc..
    """
    return xl_chr(_i + 64)


def col_offset(_col: str, _offset: int) -> str:
    """Return a column character, offset from the base by a specified amount."""
    base = xl_ord(_col)
    new = base + _offset
    return xl_chr(new)


class XlItem:
    """A single XLItem which can be written out to a specific XL Range."""

    def __init__(
        self,
        sheet_name: str,
        xl_range: str,
        write_value: xl_writable,
        input_unit: Optional[str] = None,
        target_unit: Optional[str] = None,
        range_color: Optional[Tuple[int, int, int]] = None,
        font_color: Optional[Tuple[int, int, int]] = None,
    ):
        self.sheet_name = sheet_name
        self.xl_range = xl_range
        self._write_value = write_value
        self.input_unit = input_unit
        self.target_unit = target_unit
        self.xl_range_base = xl_range
        self.range_color = range_color
        self.font_color = font_color

    @property
    def xl_row_number(self) -> int:
        return int([_ for _ in self.xl_range if _.isdigit()][0])

    @property
    def xl_col_number(self) -> int:
        return xl_ord(self.xl_range)

    @property
    def xl_col_alpha(self) -> str:
        return xl_chr(self.xl_col_number)

    @property
    def write_value(self) -> xl_writable:
        # -- Try to convert the unit (SI/IP)

        if not self.input_unit or not self.target_unit:
            return self._write_value

        if isinstance(self._write_value, (tuple, list)):
            return [converter.convert(v, self.input_unit, self.target_unit) for v in self._write_value]
        else:
            return converter.convert(self._write_value, self.input_unit, self.target_unit)

    @property
    def has_color(self) -> bool:
        """Return True if the Item has font or background color values."""
        if self.font_color and self.range_color:
            return True
        return False

    @property
    def value_is_iterable(self) -> bool:
        """Return True is the item's value is a List or Tuple"""
        return isinstance(self.write_value, (List, Tuple))

    def __str__(self):
        return f"{self.__class__.__name__}({self.sheet_name}, {self.xl_range}, {self.write_value})"

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(sheet_name={self.sheet_name}, range={self.xl_range}, value={self.write_value})"
        )

    def __len__(self) -> int:
        try:
            return len(self.write_value)  # type: ignore
        except:
            return 0


class XLItem_List:
    """A list of XLItems which yield a list when 'write_value' is called.
    This will write the data our to the columns in the row.

    ie: A1=[12, 13, 14] --> A1=12, B1=13, C1=14

    This is helpful to speed up the XL writing, since writing a single list
    is much faster than writing several XlItems one at a time.
    """

    def __init__(self, _items: List[XlItem]):
        self._items = _items

    @property
    def items(self):
        return sorted(self._items, key=lambda i: i.xl_col_number)

    def validate_xl_item_range(self, _xl_item: XlItem) -> None:
        """Raise Exception if the XlItem being added has an invalid xl_range."""
        check_item = self.items[-1]
        if _xl_item.xl_col_number != check_item.xl_col_number + 1:
            msg = (
                f"\n\tError: The XlItem with xl_range: '{_xl_item.xl_range}' "
                f"cannot be added to the XlItem_List with item "
                f"ranges: '{[i.xl_range for i in self.items]}'. "
                "Ranges should be in order ie: A, B, C, etc.."
            )
            raise Exception(msg)

    def add_new_xl_item(self, _xl_item: XlItem) -> None:
        if self.items:
            self.validate_xl_item_range(_xl_item)

        self._items.append(_xl_item)

    @property
    def write_value(self) -> List[xl_writable]:
        return [_.write_value for _ in self.items]

    @property
    def sheet_name(self) -> str:
        return self.items[0].sheet_name

    @property
    def xl_range(self) -> str:
        return self.items[0].xl_range

    @property
    def xl_row_number(self) -> int:
        return self.items[0].xl_row_number

    @property
    def xl_col_number(self) -> int:
        return self.items[0].xl_col_number

    @property
    def xl_col_alpha(self) -> str:
        return self.items[0].xl_col_alpha

    @property
    def range_color(self):
        return self.items[0].range_color

    @property
    def font_color(self):
        return self.items[0].font_color

    @property
    def has_color(self) -> bool:
        """Return True if any of the Items has font or background color values."""
        for item in self.items:
            if item.font_color or item.range_color:
                return True
        return False

    @property
    def value_is_iterable(self) -> bool:
        """Return True."""
        return True

    def __len__(self) -> int:
        return len(self._items)


def merge_xl_item_row(
    _xl_items: List[XlItem],
) -> Union[List[XlItem], List[XLItem_List], List[Union[XlItem, XLItem_List]]]:
    """Merge a List of XLItems into rows, where possible."""

    # -- Make sure all XLItems are on the same sheet
    if len({item.sheet_name for item in _xl_items}) != 1:
        return _xl_items

    # -- Make sure all the XLItems are on the same row
    if len({item.xl_row_number for item in _xl_items}) != 1:
        return _xl_items

    items_sorted = sorted(_xl_items, key=lambda i: i.xl_col_number)

    # -- Set the 'base' for each item in the set, if its xl_range
    # -- column follows the one before it. ie: "A1, B1" will get joined.
    # -- but "A1, C1" will not get joined.
    for i, item in enumerate(items_sorted):
        preceding_item = items_sorted[i - 1]
        if item.xl_col_number == preceding_item.xl_col_number + 1:
            item.xl_range_base = preceding_item.xl_range_base

    # -- Merge the items together
    d: Dict[str, Union[XlItem, XLItem_List]] = {}
    for item in items_sorted:
        if item.xl_range_base not in d.keys():
            # -- Not in the dict yet, so just
            # -- add the XlItem to the dict
            d[item.xl_range_base] = item
        else:
            existing_item = d[item.xl_range_base]

            if isinstance(existing_item, XLItem_List):
                # -- If the existing item is already a list
                # -- add the new item to it
                existing_item.add_new_xl_item(item)
                d[item.xl_range_base] = existing_item
            elif isinstance(existing_item, XlItem):
                # -- If the existing item is just an item
                # -- first, transform it into a List item
                # -- and then add the new item to the list
                existing_item = XLItem_List([existing_item])
                existing_item.add_new_xl_item(item)
                d[item.xl_range_base] = existing_item
            else:
                msg = f"\n\tError: Unsupported type: '{type(existing_item)}'"
                raise Exception(msg)

    return list(d.values())



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Valid 'types' for Foundation Options and Types"""

from enum import Enum


class CalculationSetting(Enum):
    USER_DEFINED = 6


class FoundationType(Enum):
    HEATED_BASEMENT = 1
    UNHEATED_BASEMENT = 2
    SLAB_ON_GRADE = 3
    VENTED_CRAWLSPACE = 4
    NONE = 5


class PerimeterInsulationPosition(Enum):
    UNDEFINED = 1
    HORIZONTAL = 2
    VERTICAL = 3
